# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This module is aimed to provide with tensor detail info."""
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import Streams, create_view_event_from_tensor_basic_info, ViewCommandLevelEnum


class TensorDetailInfo:
    """Manage tensor detail information."""

    def __init__(self, cache):
        self._put_command = cache.put_command
        self._metadata_stream = cache.get_stream_handler(Streams.METADATA)
        self._multi_card_tensor_stream = cache.get_stream_handler(Streams.TENSOR)
        self._multi_card_graph_stream = cache.get_stream_handler(Streams.GRAPH)
        self._multi_card_hit_stream = cache.get_stream_handler(Streams.WATCHPOINT_HIT)

    def validate_tensor_name(self, tensor_name, graph_name, rank_id):
        """
        Get the graph id of the tensor.

        Args:
            tensor_name (str): The tensor name on UI.
            graph_name (str): The graph name.
            rank_id (int): The rank id.
        """
        # validate tensor name format
        if not isinstance(tensor_name, str) or ':' not in tensor_name:
            log.error("Invalid tensor name. Received: %s", tensor_name)
            raise DebuggerParamValueError("Invalid tensor name.")
        node_name, _ = tensor_name.rsplit(':', 1)
        # check if the node name is in graph
        self._multi_card_graph_stream.get_graph_handler_by_rank_id(rank_id).validate_node_name(node_name=node_name,
                                                                                               graph_name=graph_name)

    def get_tensor_graph(self, tensor_name, graph_name, rank_id=0):
        """
        Get the graph related to specific tensor.

        Args:
            tensor_name (str): The ui name of tensor. Format like {node_name}:{slot}.
            graph_name (str): The graph name.
            rank_id (int): The rank id.

        Returns:
            dict, tensor graph, format is {'nodes': [Node object]}.
            The Node object = {
                'graph_name': <graph_name>,
                'full_name': <full node name>,
                'type': <node type>,
                'stack_info': [<stack_info>]
                'name': <node name>,
                'input': {<node name>: <Edge object>},
                'output: {<node name>: <Edge object>},
                'slots': [<Slot object>].
            }
            Edge object = {
                'data_type': <data type>,
                'edge_type': <edge type>,
                'independent_layout': bool,
                'shape': list[<dim>],
                'slot_mapping': list[pair<slot, slot>],
            }.
        """
        self.validate_tensor_name(tensor_name=tensor_name, graph_name=graph_name, rank_id=rank_id)
        graph = self._multi_card_graph_stream.get_graph_handler_by_rank_id(rank_id).get_tensor_graph(tensor_name,
                                                                                                     graph_name)
        # add watchpoint hits info and statistics info for each tensor in tensor graph.
        # record missing tensor basic info
        nodes = graph.get('graph', {}).get('nodes', [])
        missing_tensors = []
        for node in nodes:
            node['graph_name'] = graph_name
            for slot_info in node.get('slots', []):
                self._add_watchpoint_hit_info(slot_info, node, graph_name, rank_id)
                self._add_tensor_info(slot_info, node, missing_tensors, rank_id)
        # query missing tensor values from client
        self._ask_for_missing_tensor_value(missing_tensors, tensor_name, graph_name)
        return graph

    def _add_watchpoint_hit_info(self, slot_info, node, graph_name, rank_id):
        """
        Add watchpoint hit info for the tensor.

        Args:
            slot_info (dict): Slot object.
            node (dict): Node object.
            graph_name (str): Graph name.
        """
        tensor_name = ':'.join([node.get('name'), slot_info.get('slot')])
        if self._multi_card_hit_stream.check_rank_id(rank_id=rank_id):
            slot_info.update(
                self._multi_card_hit_stream.get_hit_handler_by_rank_id(rank_id).get_tensor_hit_infos(tensor_name,
                                                                                                     graph_name))

    def _add_tensor_info(self, slot_info, node, missing_tensors, rank_id):
        """
        Add the tensor info and query for missed tensors.

        Args:
            slot_info (dict): Slot object.
            node (dict): Node object.
            missing_tensors (list[TensorBasicInfo]): List of missing tensor infos.
        """
        tensor_name = ':'.join([node.get('full_name'), slot_info.get('slot')])
        node_type = node.get('type')
        tensor_info, cur_missing_tensors = self._multi_card_tensor_stream.get_tensor_handler_by_rank_id(
            rank_id).get_tensor_info_for_tensor_graph(tensor_name, node_type, self._metadata_stream.step)
        slot_info.update(tensor_info)
        if cur_missing_tensors:
            log.debug("Get missing tensor basic infos for %s", tensor_name)
            missing_tensors.extend(cur_missing_tensors)

    def _ask_for_missing_tensor_value(self, missing_tensors, tensor_name, graph_name):
        """
        Send view command to client to query for missing tensor values.

        Args:
            missing_tensors (list[TensorBasicInfo]): List of missing tensor basic infos.
            tensor_name (str): The ui name of tensor. Format like {node_name}:{slot}.
            graph_name (str): The graph name.
        """
        if not missing_tensors:
            return
        log.debug("Ask for tensor value for: %s", missing_tensors)
        view_cmd = create_view_event_from_tensor_basic_info(missing_tensors)
        self._put_command({'view_cmd': view_cmd, 'tensor_name': tensor_name, 'graph_name': graph_name, 'stats': True,
                           'level': ViewCommandLevelEnum.STATS.value})
        log.debug("Send view cmd for tensor-graphs.")

    def get_tensor_watch_points(self, tensor_name, graph_name, rank_id=0):
        """
        Get all watchpoints that the tensor hit.

        Args:
            tensor_name (str): Tensor name from UI.
            graph_name (str): The graph name.
            rank_id (int): The rank id.

        Returns:
            list, watchpoint hit infos.
        """
        # validate tensor_name
        self.validate_tensor_name(tensor_name=tensor_name, graph_name=graph_name, rank_id=rank_id)
        # get watchpoint info that the tensor hit
        if not self._multi_card_hit_stream.check_rank_id(rank_id=rank_id):
            return []
        tensor_hit_info = self._multi_card_hit_stream.get_hit_handler_by_rank_id(rank_id).get_tensor_hit_infos(
            tensor_name, graph_name)
        watch_points = tensor_hit_info.get('watch_points', [])
        return watch_points



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Histogram Processor APIs."""
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.common.validation import Validation
from mindinsight.datavisual.common.exceptions import HistogramNotExistError
from mindinsight.datavisual.processors.base_processor import BaseProcessor


class HistogramProcessor(BaseProcessor):
    """Histogram Processor."""
    def get_histograms(self, train_id, tag):
        """
        Builds a JSON-serializable object with information about histogram data.

        Args:
            train_id (str): The ID of the events data.
            tag (str): The name of the tag the histogram data all belong to.

        Returns:
            dict, a dict including the `train_id`, `tag`, and `histograms'.
                    {
                        "train_id": ****,
                        "tag": ****,
                        "histograms": [{
                            "wall_time": ****,
                            "step": ****,
                            "bucket": [[**, **, **]],
                            },
                            {...}
                        ]
                    }
        """
        Validation.check_param_empty(train_id=train_id, tag=tag)
        logger.info("Start to process histogram data...")
        try:
            tensors = self._data_manager.list_tensors(train_id, tag)
        except ParamValueError as err:
            raise HistogramNotExistError(err.message)

        histograms = []
        for tensor in tensors:
            histogram = tensor.value
            buckets = histogram.buckets()
            histograms.append({
                "wall_time": tensor.wall_time,
                "step": tensor.step,
                "buckets": buckets
            })

        logger.info("Histogram data processing is finished!")
        response = {
            "train_id": train_id,
            "tag": tag,
            "histograms": histograms
        }
        return response



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define schema of model lineage input parameters."""
from marshmallow import Schema, fields, ValidationError, pre_load, validates

from mindinsight.lineagemgr.common.exceptions.error_code import LineageErrorMsg, LineageErrors
from mindinsight.lineagemgr.common.exceptions.exceptions import LineageParamTypeError, LineageParamValueError
from mindinsight.lineagemgr.common.utils import enum_to_list
from mindinsight.lineagemgr.querier.querier import LineageType
from mindinsight.lineagemgr.querier.query_model import FIELD_MAPPING
from mindinsight.utils.exceptions import MindInsightException


class SearchModelConditionParameter(Schema):
    """Define the search model condition parameter schema."""
    summary_dir = fields.Dict()
    loss_function = fields.Dict()
    train_dataset_path = fields.Dict()
    train_dataset_count = fields.Dict()
    test_dataset_path = fields.Dict()
    test_dataset_count = fields.Dict()
    network = fields.Dict()
    optimizer = fields.Dict()
    learning_rate = fields.Dict()
    epoch = fields.Dict()
    batch_size = fields.Dict()
    device_num = fields.Dict()
    loss = fields.Dict()
    model_size = fields.Dict()
    limit = fields.Int(validate=lambda n: 0 < n <= 100)
    offset = fields.Int(validate=lambda n: 0 <= n <= 100000)
    sorted_name = fields.Str()
    sorted_type = fields.Str(allow_none=True)
    dataset_mark = fields.Dict()
    lineage_type = fields.Dict()

    @staticmethod
    def check_dict_value_type(data, value_type):
        """Check dict value type and int scope."""
        for key, value in data.items():
            if key in ["in", "not_in"]:
                if not isinstance(value, (list, tuple)):
                    raise ValidationError("The value of `in` operation must be list or tuple.")
            else:
                if not isinstance(value, value_type):
                    raise ValidationError("Wrong value type.")
                if value_type is int:
                    if value < 0 or value > pow(2, 63) - 1:
                        raise ValidationError("Int value should <= pow(2, 63) - 1.")
                    if isinstance(value, bool):
                        raise ValidationError("Wrong value type.")

    @staticmethod
    def check_param_value_type(data):
        """Check input param's value type."""
        for key, value in data.items():
            if key == "in":
                if not isinstance(value, (list, tuple)):
                    raise ValidationError("The value of `in` operation must be list or tuple.")
            else:
                if isinstance(value, bool) or \
                        (not isinstance(value, float) and not isinstance(value, int)):
                    raise ValidationError("Wrong value type.")

    @staticmethod
    def check_operation(data):
        """Check input param's compare operation."""
        if not set(data.keys()).issubset(['in', 'eq', 'not_in']):
            raise ValidationError("Its operation should be `eq`, `in` or `not_in`.")

    @validates("loss")
    def check_loss(self, data):
        """Check loss."""
        SearchModelConditionParameter.check_param_value_type(data)

    @validates("learning_rate")
    def check_learning_rate(self, data):
        """Check learning_rate."""
        SearchModelConditionParameter.check_param_value_type(data)

    @validates("loss_function")
    def check_loss_function(self, data):
        """Check loss function."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("train_dataset_path")
    def check_train_dataset_path(self, data):
        """Check train dataset path."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("train_dataset_count")
    def check_train_dataset_count(self, data):
        """Check train dataset count."""
        SearchModelConditionParameter.check_dict_value_type(data, int)

    @validates("test_dataset_path")
    def check_test_dataset_path(self, data):
        """Check test dataset path."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("test_dataset_count")
    def check_test_dataset_count(self, data):
        """Check test dataset count."""
        SearchModelConditionParameter.check_dict_value_type(data, int)

    @validates("network")
    def check_network(self, data):
        """Check network."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("optimizer")
    def check_optimizer(self, data):
        """Check optimizer."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("epoch")
    def check_epoch(self, data):
        """Check epoch."""
        SearchModelConditionParameter.check_dict_value_type(data, int)

    @validates("batch_size")
    def check_batch_size(self, data):
        """Check batch size."""
        SearchModelConditionParameter.check_dict_value_type(data, int)

    @validates("device_num")
    def check_device_num(self, data):
        """Check device num."""
        SearchModelConditionParameter.check_dict_value_type(data, int)

    @validates("model_size")
    def check_model_size(self, data):
        """Check model size."""
        SearchModelConditionParameter.check_dict_value_type(data, int)

    @validates("summary_dir")
    def check_summary_dir(self, data):
        """Check summary dir."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("dataset_mark")
    def check_dataset_mark(self, data):
        """Check dataset mark."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)

    @validates("lineage_type")
    def check_lineage_type(self, data):
        """Check lineage type."""
        SearchModelConditionParameter.check_operation(data)
        SearchModelConditionParameter.check_dict_value_type(data, str)
        recv_types = []
        for key, value in data.items():
            if key == "in":
                recv_types = value
            else:
                recv_types.append(value)

        lineage_types = enum_to_list(LineageType)
        if not set(recv_types).issubset(lineage_types):
            raise ValidationError("Given lineage type should be one of %s." % lineage_types)

    @pre_load
    def check_comparison(self, data, **kwargs):
        """Check comparison for all parameters in schema."""
        for attr, condition in data.items():
            if attr in ["limit", "offset", "sorted_name", "sorted_type", 'lineage_type']:
                continue

            if not isinstance(attr, str):
                raise LineageParamValueError('The search attribute not supported.')

            if attr not in FIELD_MAPPING and not attr.startswith(('metric/', 'user_defined/')):
                raise LineageParamValueError('The search attribute not supported.')

            if not isinstance(condition, dict):
                raise LineageParamTypeError("The search_condition element {} should be dict."
                                            .format(attr))

            for key in condition.keys():
                if key not in ["eq", "lt", "gt", "le", "ge", "in", "not_in"]:
                    raise LineageParamValueError("The compare condition should be in "
                                                 "('eq', 'lt', 'gt', 'le', 'ge', 'in', 'not_in').")

            if attr.startswith('metric/'):
                if len(attr) == 7:
                    raise LineageParamValueError(
                        'The search attribute not supported.'
                    )
                try:
                    SearchModelConditionParameter.check_param_value_type(condition)
                except ValidationError:
                    raise MindInsightException(
                        error=LineageErrors.LINEAGE_PARAM_METRIC_ERROR,
                        message=LineageErrorMsg.LINEAGE_METRIC_ERROR.value.format(attr)
                    )
        return data



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This file is used to define the node of graph and associated base types.
"""
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.datavisual.common.log import logger as log


class NodeTree:
    """A class for building a node tree."""
    def __init__(self, node_name='', node_type=None, is_dynamic_shape_node=False):
        self.node_name = node_name
        self._node_type = node_type
        self.is_dynamic_shape_node = is_dynamic_shape_node
        self._children = {}

    @property
    def node_type(self):
        """The property of node type."""
        return self._node_type

    @node_type.setter
    def node_type(self, value):
        """Set the node type."""
        self._node_type = value

    def add(self, name, node_type=None, is_dynamic_shape_node=False):
        """Add sub node."""
        sub_name = '/'.join([self.node_name, name]) if self.node_name else name
        sub_node = NodeTree(sub_name, node_type, is_dynamic_shape_node)
        self._children[name] = sub_node
        return sub_node

    def get(self, sub_name):
        """Get sub node."""
        return self._children.get(sub_name)

    def get_children(self):
        """Get all childrens."""
        for name_scope, sub_node in self._children.items():
            yield name_scope, sub_node

    def remove(self, sub_name):
        """Remove sub node."""
        try:
            self._children.pop(sub_name)
        except KeyError as err:
            log.error("Failed to find node %s. %s", sub_name, err)
            raise ParamValueError("Failed to find node {}".format(sub_name))



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Node in the computational graph."""
import collections
from abc import ABC

from mindinsight.debugger.api.debugger_tensor import DebuggerTensorImpl
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.utils import validate_slots, parse_param_to_iterable_obj
from mindinsight.domain.graph.base import NodeType

NodeFeature = collections.namedtuple('NodeFeature', ['name', 'rank', 'stack',
                                                     'graph_name', 'root_graph_id'])
NodeUniqueId = collections.namedtuple('NodeUniqueId', ['name', 'rank', 'graph_name'])


class Node(ABC):
    """
    Node in the computational graph.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        node_feature (namedtuple): The node feature, including the following information:

            - name (str): The node name.
            - rank (int): The rank id.
            - stack (iterable[dict]): The stack information. The format of each item is like:

              .. code-block::

                  {
                      'file_path': str,
                      'line_no': int,
                      'code_line': str
                  }

            - graph_name (str): The graph name.
            - root_graph_id (int): The root graph id.

    Note:
        - Users should not instantiate this class manually.
        - The instances of this class is immutable.
    """

    def __init__(self, node_feature):
        self._node_feature = node_feature

    @property
    def name(self):
        """
        Get the full name of this node.

        Returns:
            str, the full name of the node.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node = list(my_run.select_nodes("conv"))[0]
                >>> print(node.name)
                conv1.weight
        """
        return self._node_feature.name

    @property
    def stack(self):
        """
        Get stack info of the node.

        Returns:
            iterable[dict], each item format as follows,

            .. code-block::

                {
                    'file_path': str,
                    'line_no': int,
                    'code_line': str
                }

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node = list(my_run.select_nodes("Conv2D-op13"))[0]
                >>> # print(node.stack)
                >>> # the print result is as follows
                >>> # [{'file_path': '/path', 'line_no': 266, 'code_line': 'output = self.conv2d(x, self.weight)',
                >>> # 'has_substack': False},
                >>> # {'file_path': '/path', 'line_no': 55, 'code_line': 'x = self.conv2(x), 'has_substack': False}]
        """
        return self._node_feature.stack

    @property
    def rank(self) -> int:
        """
        Get rank info.

        Returns:
            int, the rank id to which the node belong.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node = list(my_run.select_nodes("conv"))[0]
                >>> print(node.rank)
                0
        """
        return self._node_feature.rank

    @property
    def root_graph_id(self) -> int:
        """
        Get the root graph id to which the dumped tensor of current node will belong.

        Returns:
            int, the root graph id.
        """
        return self._node_feature.root_graph_id

    @property
    def graph_name(self) -> str:
        """
        Get graph name of current node.

        Returns:
            str, the graph name.
        """
        return self._node_feature.graph_name

    def get_input_tensors(
            self,
            iterations=None,
            slots=None):
        """
        Get the input tensors of the node.

        Args:
            iterations (Iterable[int], optional): The iterations to which the returned
                tensor should belong. ``None`` means all available iterations will be considered.
                Default: ``None``.
            slots (Iterable[int], optional): The slots in which the returned tensors
                should be. ``None`` means all available slots will be considered. Default: ``None``.

        Returns:
            Iterable[DebuggerTensor], the input tensors of the node.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node = list(my_run.select_nodes("Conv2D-op13"))[0]
                >>> input_tensors = node.get_input_tensors(iterations=[0], slots=[0])
        """

    def get_output_tensors(
            self,
            iterations=None,
            slots=None):
        """
        Get the output tensors of this node.

        Args:
            iterations (Iterable[int], optional): The iterations to which the returned
                tensor should belong. ``None`` means all available iterations will be considered.
                Default: ``None``.
            slots (Iterable[int], optional): The slots in which the returned tensors
                should be. ``None`` means all available slots will be considered. Default: ``None``.

        Returns:
            Iterable[DebuggerTensor], the output tensors of the node.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node = list(my_run.select_nodes("Conv2D-op13"))[0]
                >>> output_tensors = node.get_output_tensors(iterations=[0], slots=[0])
        """

    def __str__(self):
        feature = f"rank: {self.rank}\n" \
                  f"graph_name: {self.graph_name}\n" \
                  f"node_name: {self.name}"
        return feature


class NodeImpl(Node):
    """Node implementation."""

    def __init__(self, base_node, node_type):
        self._base_node = base_node
        self._node_type = node_type
        node_feature = self._get_node_feature()
        super(NodeImpl, self).__init__(node_feature)
        self.debugger_engine = None
        self.input_nodes = []
        self.downstream = []

    @property
    def name(self):
        """
        Get the full name of this node.

        Returns:
            str, the full name of the node.

        Examples:
            >>> from mindinsight.debugger import DumpAnalyzer
            >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
            >>> node = list(my_run.select_nodes("conv"))[0]
            >>> print(node.name)
            conv1.weight
        """
        if hasattr(self._base_node, 'type') and (self._base_node.type == "Load" or "Summary" in self._base_node.type):
            name = self._base_node.name.split('/')[-1]
            node_id = name.split('op')[-1]
            name = f'{self._base_node.type}-op{node_id}'
            return f'{self._base_node.scope}/{name}'
        return self._node_feature.name

    @property
    def node_type(self):
        """Get the node type."""
        return self._node_type

    @property
    def base_node(self):
        """Get the base node."""
        return self._base_node

    @property
    def output_slot_size(self):
        """Get the base node."""
        return self._base_node.output.slot_size

    @property
    def full_name_with_graph(self):
        """Return node name startswith graph name."""
        return f"{self.graph_name}/{self.name}"

    @property
    def unique_id(self):
        """Get the unique id of the node."""
        return NodeUniqueId(self.name, self.rank, self.graph_name)

    def _get_node_feature(self):
        """Get node feature."""
        node = self._base_node
        if self._node_type in (NodeType.CONSTANT, NodeType.PARAMETER):
            feature = NodeFeature(name=node.name,
                                  rank=node.rank_id,
                                  stack=[],
                                  graph_name=node.graph_name,
                                  root_graph_id=node.root_graph_id)
        elif self._node_type == NodeType.OPERATOR:
            feature = NodeFeature(name=node.full_name,
                                  rank=node.rank_id,
                                  stack=node.stack,
                                  graph_name=node.graph_name,
                                  root_graph_id=node.root_graph_id)
        else:
            raise DebuggerParamValueError(f"Invalid node_type, got {self._node_type}")
        return feature

    def get_input_tensors(
            self,
            iterations=None,
            slots=None):
        """
        Get the input tensors of the node.

        Args:
            iterations (Iterable[int], optional): The iterations to which the returned
                tensor should belong. ``None`` means all available iterations will be considered.
                Default: ``None``.
            slots (Iterable[int], optional): The slots in which the returned tensors
                should be. ``None`` means all available slots will be considered. Default: ``None``.

        Returns:
            Iterable[DebuggerTensor], the input tensors of the node.
        """
        validate_slots(slots)
        iterations = self._get_iterable_iterations(iterations)
        current_index = 0
        res = []
        for node in self.input_nodes:
            if slots is not None:
                query_slots = [slot for slot in range(node.output_slot_size) if slot + current_index in slots]
            else:
                query_slots = None
            res.extend(node.get_output_tensors(iterations, query_slots))
            current_index += node.output_slot_size
        return res

    def get_output_tensors(
            self,
            iterations=None,
            slots=None):
        """
        Get the output tensors of this node.

        Args:
            iterations (Iterable[int]): The iterations to which the returned
                tensor should belong. Default: ``None``.
            slots (Iterable[int]): The slots in which the returned tensors
                should be. Default: ``None``.

        Returns:
            Iterable[DebuggerTensor], the output tensors of the node.
        """
        validate_slots(slots)
        iterations = self._get_iterable_iterations(iterations)
        output_slots = list(range(self.output_slot_size))
        if slots is None:
            query_slots = output_slots
        else:
            query_slots = [slot for slot in output_slots if slot in slots]
        res = []
        for slot in query_slots:
            for iteration in iterations:
                res.append(DebuggerTensorImpl(self.copy(), slot, iteration))
        return res

    def _get_iterable_iterations(self, iterations):
        """
        Validate input and return iterable iterations.

        Args:
           iterations (Union[int, list[int], None], optional): The specified iterations.

        Returns:
            list[int], list of iteration.
        """
        data_loader = self.debugger_engine.data_loader
        total_dumped_steps = data_loader.load_dumped_step(self.rank)
        dumped_iterations = total_dumped_steps.get(self.rank, {}).get(self.root_graph_id, [])
        iterations = parse_param_to_iterable_obj(iterations, 'iterations', dumped_iterations, False)
        return iterations

    def copy(self):
        """Copy Node."""
        node = NodeImpl(self._base_node, self._node_type)
        node.debugger_engine = self.debugger_engine
        node.input_nodes = self.input_nodes.copy()
        node.downstream = self.downstream.copy()
        return node



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Histogram data container."""
from mindinsight.datavisual.data_transform.histogram import Histogram, Bucket, mask_invalid_number
from mindinsight.datavisual.proto_files.mindinsight_summary_pb2 import Summary


class HistogramContainer:
    """
        Histogram data container.

        Args:
            histogram_message (Summary.Histogram): Histogram message in summary file.
    """

    def __init__(self, histogram_message: Summary.Histogram):
        original_buckets = [Bucket(bucket.left, bucket.width, bucket.count) for bucket in histogram_message.buckets]
        # Ensure buckets are sorted from min to max.
        original_buckets.sort(key=lambda bucket: bucket.left)
        self._count = sum(bucket.count for bucket in original_buckets)
        self._max = mask_invalid_number(histogram_message.max)
        self._min = mask_invalid_number(histogram_message.min)
        self._histogram = Histogram(tuple(original_buckets), self._max, self._min, self._count)

    @property
    def max(self):
        """Gets max value of the tensor."""
        return self._max

    @property
    def min(self):
        """Gets min value of the tensor."""
        return self._min

    @property
    def count(self):
        """Gets valid number count of the tensor."""
        return self._count

    @property
    def histogram(self):
        """Gets histogram data"""
        return self._histogram

    def buckets(self):
        """Gets histogram buckets"""
        return self._histogram.buckets()



# Copyright 2019-2021 Huawei Technologies Co., Ltd.All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Setup."""

import sys
import os
import shutil
import stat
import platform
import shlex
import subprocess
import types
from importlib import import_module
from setuptools import setup
from setuptools.command.egg_info import egg_info
from setuptools.command.build_py import build_py
from setuptools.command.install import install


def get_readme_content():
    pwd = os.path.dirname(os.path.realpath(__file__))
    with open(os.path.join(pwd, 'README.md'), encoding='UTF-8') as f:
        return f.read()


def get_version():
    """
    Get version.

    Returns:
        str, mindinsight version.
    """
    machinery = import_module('importlib.machinery')
    module_path = os.path.join(os.path.dirname(__file__), 'mindinsight', '_version.py')
    module_name = '__mindinsightversion__'

    version_module = types.ModuleType(module_name)
    loader = machinery.SourceFileLoader(module_name, module_path)
    loader.exec_module(version_module)
    return version_module.VERSION


def get_platform():
    """
    Get platform name.

    Returns:
        str, platform name in lowercase.
    """
    return platform.system().strip().lower()


def get_description():
    """
    Get description.

    Returns:
        str, wheel package description.
    """
    os_info = get_platform()
    cpu_info = platform.machine().strip()

    cmd = "git log --format='[sha1]:%h, [branch]:%d' -1"
    process = subprocess.Popen(
        shlex.split(cmd),
        shell=False,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    stdout, _ = process.communicate()
    if not process.returncode:
        git_version = stdout.decode().strip()
        return 'mindinsight platform: %s, cpu: %s, git version: %s' % (os_info, cpu_info, git_version)

    return 'mindinsight platform: %s, cpu: %s' % (os_info, cpu_info)


def get_install_requires():
    """
    Get install requirements.

    Returns:
        list, list of dependent packages.
    """
    with open('requirements.txt') as file:
        return file.read().strip().splitlines()


def update_permissions(path):
    """
    Update permissions.

    Args:
        path (str): Target directory path.
    """
    for dirpath, dirnames, filenames in os.walk(path):
        for dirname in dirnames:
            dir_fullpath = os.path.join(dirpath, dirname)
            os.chmod(dir_fullpath, stat.S_IREAD | stat.S_IWRITE | stat.S_IEXEC | stat.S_IRGRP | stat.S_IXGRP)
        for filename in filenames:
            file_fullpath = os.path.join(dirpath, filename)
            os.chmod(file_fullpath, stat.S_IREAD)


def run_script(script):
    """
    Run script.

    Args:
        script (str): Target script file path.

    Returns:
        int, return code.
    """
    cmd = '/bin/bash {}'.format(script)
    process = subprocess.Popen(
        shlex.split(cmd),
        shell=False
    )
    process.communicate()
    return process.returncode


def build_dependencies():
    """Build dependencies."""
    build_dir = os.path.join(os.path.dirname(__file__), 'build')

    sys.stdout.write('building ui ...\n')
    ui_script = os.path.join(build_dir, 'scripts', 'ui.sh')
    rc = run_script(ui_script)
    if rc:
        sys.stdout.write('building ui failure\n')
        sys.exit(1)


class EggInfo(egg_info):
    """Egg info."""

    def run(self):
        build_dependencies()
        egg_info_dir = os.path.join(os.path.dirname(__file__), 'mindinsight.egg-info')
        shutil.rmtree(egg_info_dir, ignore_errors=True)
        super().run()
        update_permissions(egg_info_dir)


class BuildPy(build_py):
    """Build py files."""

    def run(self):
        mindinsight_lib_dir = os.path.join(os.path.dirname(__file__), 'build', 'lib', 'mindinsight')
        shutil.rmtree(mindinsight_lib_dir, ignore_errors=True)
        super().run()
        update_permissions(mindinsight_lib_dir)


class Install(install):
    """Install."""

    def run(self):
        super().run()
        if sys.argv[-1] == 'install':
            pip = import_module('pip')
            mindinsight_dir = os.path.join(os.path.dirname(pip.__path__[0]), 'mindinsight')
            update_permissions(mindinsight_dir)


if __name__ == '__main__':
    version_info = sys.version_info
    if (version_info.major, version_info.minor) < (3, 7):
        sys.stderr.write('Python version should be at least 3.7\r\n')
        sys.exit(1)

    setup(
        name='mindinsight',
        version=get_version(),
        author='The MindSpore Authors',
        author_email='contact@mindspore.cn',
        url='https://www.mindspore.cn',
        download_url='https://gitee.com/mindspore/mindinsight/tags',
        project_urls={
            'Sources': 'https://gitee.com/mindspore/mindinsight',
            'Issue Tracker': 'https://gitee.com/mindspore/mindinsight/issues',
        },
        description=get_description(),
        long_description=get_readme_content(),
        long_description_content_type="text/markdown",
        packages=['mindinsight'],
        platforms=[get_platform()],
        include_package_data=True,
        cmdclass={
            'egg_info': EggInfo,
            'build_py': BuildPy,
            'install': Install,
        },
        entry_points={
            'console_scripts': [
                'mindinsight=mindinsight.utils.command:main',
                'mindwizard=mindinsight.wizard.cli:cli_entry',
                'mindoptimizer=mindinsight.optimizer.cli:cli_entry',
            ],
        },
        python_requires='>=3.7',
        install_requires=get_install_requires(),
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Environment :: Web Environment',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        license='Apache 2.0',
        keywords='mindinsight',
    )



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Transformer."""
from mindinsight.optimizer.utils.param_handler import match_value_type
from mindinsight.optimizer.common.enums import HyperParamSource, HyperParamKey


class Transformer:
    """Transformer."""
    @staticmethod
    def transform_list_to_dict(params_info, suggest_list):
        """Transform from tuner."""
        suggest_list = match_value_type(suggest_list, params_info)
        param_dict = {}
        user_defined_info = {}
        for index, param_name in enumerate(params_info):
            param_item = {param_name: suggest_list[index]}
            param_dict.update(param_item)
            source = params_info.get(param_name).get(HyperParamKey.SOURCE.value)
            if source is not None and source == HyperParamSource.USER_DEFINED.value:
                user_defined_info.update(param_item)

        return param_dict, user_defined_info



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The analyser module."""
from . import analyser, minddata_pipeline_analyser, step_trace_analyser, minddata_analyser, \
    timeline_analyser, msprof_timeline_analyser, cpu_analyser, gpu_analyser, memory_usage_analyser, \
    minddata_cpu_utilization_analyser, cluster_analyser, flops_analyser, parallel_strategy_analyser, parallel_analyser



# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""alexnet module."""
from mindinsight.wizard.network.generic_network import GenericNetwork


class Network(GenericNetwork):
    """Network code generator."""
    name = 'alexnet'
    supported_datasets = ['Cifar10', 'ImageNet']
    supported_loss_functions = ['SoftmaxCrossEntropyWithLogits']
    supported_optimizers = ['Momentum', 'Adam', 'SGD']



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utils for params."""
import math
import numpy as np

from mindinsight.lineagemgr.model import LineageTable, USER_DEFINED_PREFIX, METRIC_PREFIX
from mindinsight.optimizer.common.enums import HyperParamKey, HyperParamType, HyperParamSource, TargetKey, \
    TargetGoal, TunableSystemDefinedParams, TargetGroup, SystemDefinedTargets
from mindinsight.optimizer.common.log import logger


def generate_param(param_info, n=1):
    """Generate param."""
    value = None
    if HyperParamKey.BOUND.value in param_info:
        bound = param_info[HyperParamKey.BOUND.value]
        value = np.random.uniform(bound[0], bound[1], n)
        if param_info[HyperParamKey.TYPE.value] == HyperParamType.INT.value:
            value = value.astype(HyperParamType.INT.value)
    if HyperParamKey.CHOICE.value in param_info:
        indexes = np.random.randint(0, len(param_info[HyperParamKey.CHOICE.value]), n)
        value = [param_info[HyperParamKey.CHOICE.value][index] for index in indexes]
    if HyperParamKey.DECIMAL.value in param_info:
        value = np.around(value, decimals=param_info[HyperParamKey.DECIMAL.value])
    return np.array(value)


def generate_arrays(params_info: dict, n=1):
    """Generate arrays."""
    suggest_params = None
    for _, param_info in params_info.items():
        suggest_param = generate_param(param_info, n).reshape((-1, 1))
        if suggest_params is None:
            suggest_params = suggest_param
        else:
            suggest_params = np.hstack((suggest_params, suggest_param))
    if n == 1:
        return suggest_params[0]
    return suggest_params


def match_value_type(array, params_info: dict):
    """Make array match params type."""
    array_new = []
    index = 0
    for _, param_info in params_info.items():
        value = array[index]
        bound = param_info.get(HyperParamKey.BOUND.value)
        choice = param_info.get(HyperParamKey.CHOICE.value)
        if bound is not None:
            value = max(bound[0], array[index])
            value = min(bound[1], value)
        if choice is not None:
            nearest_index = int(np.argmin(np.fabs(np.array(choice) - value)))
            value = choice[nearest_index]
        if param_info.get(HyperParamKey.TYPE.value) == HyperParamType.INT.value:
            value = int(value)
            if bound is not None and value < bound[0]:
                value = math.ceil(bound[0])
            elif bound is not None and value >= bound[1]:
                # bound[1] is 2.0, value is 1; bound[1] is 2.1, value is 2
                value = math.floor(bound[1]) - 1
        if HyperParamKey.DECIMAL.value in param_info:
            value = np.around(value, decimals=param_info[HyperParamKey.DECIMAL.value])
        array_new.append(value)
        index += 1
    return array_new


def organize_params_target(lineage_table: LineageTable, params_info: dict, target_info):
    """Organize params and target."""
    empty_result = np.array([])
    if lineage_table is None:
        return empty_result, empty_result

    param_keys = []
    for param_key, param_info in params_info.items():
        # It will be a user_defined param:
        # 1. if 'source' is specified as 'user_defined'
        # 2. if 'source' is not specified and the param is not a system_defined key
        source = param_info.get(HyperParamKey.SOURCE.value)
        prefix = _get_prefix(param_key, source, HyperParamSource.USER_DEFINED.value,
                             USER_DEFINED_PREFIX, TunableSystemDefinedParams.list_members())
        param_key = f'{prefix}{param_key}'
        if prefix == USER_DEFINED_PREFIX:
            param_info[HyperParamKey.SOURCE.value] = HyperParamSource.USER_DEFINED.value
        else:
            param_info[HyperParamKey.SOURCE.value] = HyperParamSource.SYSTEM_DEFINED.value

        param_keys.append(param_key)

    target_name = target_info[TargetKey.NAME.value]
    group = target_info.get(TargetKey.GROUP.value)
    prefix = _get_prefix(target_name, group, TargetGroup.METRIC.value,
                         METRIC_PREFIX, SystemDefinedTargets.list_members())
    target_name = prefix + target_name
    lineage_df = lineage_table.dataframe_data
    try:
        lineage_df = lineage_df[param_keys + [target_name]]
        lineage_df = lineage_df.dropna(axis=0, how='any')

        target_column = np.array(lineage_df[target_name])
        if TargetKey.GOAL.value in target_info and \
                target_info.get(TargetKey.GOAL.value) == TargetGoal.MAXIMUM.value:
            target_column = -target_column

        return np.array(lineage_df[param_keys]), target_column
    except KeyError as exc:
        logger.warning("Some keys not exist in specified params or target. It will suggest params randomly."
                       "Detail: %s.", str(exc))
    return empty_result, empty_result


def _get_prefix(name, field, other_defined_field, other_defined_prefix, system_defined_fields):
    if (field == other_defined_field) or (field is None and name not in system_defined_fields):
        return other_defined_prefix
    return ''



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Base writer."""
from abc import abstractmethod


class Writer:
    """Base writer for writers."""
    @abstractmethod
    def add(self, value):
        """
        Abstract method for adding value.

        Args:
            value (object): scalar, tensor or image value with wall_time, tag and step.
        """

    @abstractmethod
    def write(self):
        """Abstract method for writing file."""



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The base analyser."""
import functools
from abc import ABC, abstractmethod

from marshmallow import ValidationError

from mindinsight.profiler.common.exceptions.exceptions import \
    ProfilerColumnNotExistException, ProfilerPathErrorException, \
    ProfilerIOException, ProfilerColumnNotSupportSortException
from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.validator.validate_path import \
    validate_and_normalize_path


class BaseAnalyser(ABC):
    """
    The base analyser.

    A concrete analyser class can be constructed by inheriting the class. The
    analyser provides the ability to filter, sort and group. The subclass only
    need to implement `_load`, `_filter`, `_sort` and `_group`. The condition
    defines the rules for filtering, sorting and grouping.

    Args:
        profiling_dir (str): The directory where the parsed profiling files
            are located.
        device_id (str): The device ID.

    Raises:
        ProfilerPathErrorException: If the profiling dir is invalid.
    """
    _col_names = []

    def __init__(self, profiling_dir, device_id):
        self._profiling_dir = self._normalize_profiling_dir(profiling_dir)
        self._device_id = device_id
        self._data = []
        self._result = None
        self._display_col_names = None
        self._size = 0
        self._none_filter_condition_key = []
        self._none_sort_col_names = []
        self._all_type = []
        self._ms_to_us = 1e3
        self._round_digits = 6
        self._ms_round_digits = 3

        try:
            self._load()
        except IOError as err:
            logger.exception(err)
            raise ProfilerIOException()

    @property
    def col_names(self):
        """The column names in the parsed profiling file."""
        return self._col_names

    @property
    def data(self):
        """The data in the parsed profiling file."""
        return self._data

    def query(self, condition=None):
        """
        Query data according to the condition.

        Args:
            condition (dict): The search condition, including filter condition,
                sort condition and group condition. If the condition is `None`,
                all data will be returned. Default: None.

        Returns:
            dict, the result after filtered, sorted and grouped.
        """
        if condition is None:
            condition = {}
        filter_condition = condition.get('filter_condition', {})
        sort_condition = condition.get('sort_condition')
        group_condition = condition.get('group_condition')

        self._result = []
        self._display_col_names = self._col_names[:]
        self._filter(filter_condition)
        self._size = len(self._result)
        if sort_condition:
            self._sort(sort_condition)
        if group_condition:
            self._group(group_condition)
        return self._organize_query_result()

    @abstractmethod
    def _load(self):
        """Load data according to the parsed profiling files."""

    @abstractmethod
    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

    def _sort(self, sort_condition: dict):
        """
        Sort the profiling data according to the filter condition.

        Args:
            sort_condition (dict): The sort condition.

        Raises:
            ProfilerColumnNotExistException: If the sort name does not exist.
        """
        def _cmp(item1, item2):
            value1 = item1[index]
            value2 = item2[index]

            if value1 is None and value2 is None:
                cmp_result = 0
            elif value1 is None:
                cmp_result = -1
            elif value2 is None:
                cmp_result = 1
            else:
                try:
                    cmp_result = (value1 > value2) - (value1 < value2)
                except TypeError:
                    type1 = type(value1).__name__
                    type2 = type(value2).__name__
                    cmp_result = (type1 > type2) - (type1 < type2)
            return cmp_result

        sort_name = sort_condition.get('name')
        sort_type = sort_condition.get('type', 'descending')
        reverse = sort_type == 'descending'
        if not sort_name:
            return
        try:
            index = self._col_names.index(sort_name)
        except ValueError:
            raise ProfilerColumnNotExistException(sort_name)
        if self._none_sort_col_names and sort_name in self._none_sort_col_names:
            raise ProfilerColumnNotSupportSortException(sort_name)
        self._result.sort(key=functools.cmp_to_key(_cmp), reverse=reverse)

    def _group(self, group_condition: dict):
        """
        Group the profiling data according to the group condition.

        Args:
            group_condition (dict): The group condition.
        """
        limit = group_condition.get('limit')
        offset = group_condition.get('offset')
        if limit is None and offset is None:
            return
        if limit is None:
            limit = 10
        if offset is None:
            offset = 0
        self._result = self._result[limit * offset: limit * (offset + 1)]

    def _default_filter(self, item, condition):
        """
        The default filter method.

        Args:
            item (list[Union[str, float, int]]): A piece of data to be filtered.
            condition (dict): The filter condition.

        Returns:
            bool, `True` if the item is satisfied.
        """
        for condition_key, condition_value in condition.items():
            if condition_key in self._none_filter_condition_key:
                continue
            if condition_key in self._col_names:
                index = self._col_names.index(condition_key)
                actual_value = item[index]
                if isinstance(condition_value, dict):
                    for exp_key, exp_value in condition_value.items():
                        if not self._is_match_condition(
                                exp_key, exp_value, actual_value):
                            return False
        return True

    def _is_match_condition(self, exp_key, exp_value, actual_value):
        """
        Check whether the actual value meets the expect condition.

        Args:
            exp_key (str): Expect key of the condition.
            exp_value (str): Expect value.
            actual_value (str): Actual value.

        Returns:
            bool, `True` if the actual meets the expect condition, else `False`.
        """
        if exp_key == 'in':
            if actual_value not in exp_value:
                return False
        elif exp_key == 'not_in':
            if actual_value in exp_value:
                return False
        elif exp_key == 'partial_match_str_in':
            for partial_match_str in exp_value:
                if partial_match_str.lower() in actual_value.lower():
                    return True
            return False
        else:
            return False

        return True

    def _normalize_profiling_dir(self, profiling_dir):
        """
        Normalize the profiling dir.

        Args:
            profiling_dir (str): The directory where the parsed profiling files
                are located.

        Returns:
            str, the normalized profiling dir.
        """
        try:
            normalized_profiling_dir = validate_and_normalize_path(
                profiling_dir, 'profiler'
            )
        except ValidationError:
            raise ProfilerPathErrorException('The profiling dir is invalid.')
        return normalized_profiling_dir

    def _organize_query_result(self):
        """
        Organize the query result.

        Returns:
            dict, the query result.
        """
        return {
            'col_name': self._display_col_names,
            'object': self._result,
            'size': self._size
        }

    def _format_float_data(self, float_data):
        """Format the float data."""
        if float_data == '-':
            return float_data
        if isinstance(float_data, str):
            float_data = float(float_data)

        if float_data < 1:
            float_data = round(float_data, self._round_digits)
        else:
            float_data = round(float_data, self._ms_round_digits)
        return float_data



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""BaseDataset module."""
import os

from mindinsight.wizard.base.dataset import BaseDataset
from mindinsight.wizard.base.templates import TemplateManager
from mindinsight.wizard.conf.constants import TEMPLATES_BASE_DIR


class Dataset(BaseDataset):
    """BaseDataset code generator."""
    name = 'MNIST'

    def __init__(self):
        super(Dataset, self).__init__()
        self._network = None
        self.template_manager = None

    def set_network(self, network_maker):
        self._network = network_maker
        template_dir = os.path.join(TEMPLATES_BASE_DIR,
                                    'network',
                                    network_maker.name.lower(),
                                    'dataset',
                                    self.name.lower())
        self.template_manager = TemplateManager(template_dir)

    def configure(self):
        """Configure the network options."""
        return self.settings

    def generate(self, **options):
        source_files = self.template_manager.render(**options)
        return source_files



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define custom exception."""

from mindinsight.utils.constant import OptimizerErrors
from mindinsight.utils.exceptions import MindInsightException


class SamplesNotEnoughError(MindInsightException):
    """Param importance calculated error."""
    def __init__(self, error_msg="Param importance calculated error."):
        super(SamplesNotEnoughError, self).__init__(OptimizerErrors.SAMPLES_NOT_ENOUGH,
                                                    error_msg,
                                                    http_code=400)


class CorrelationNanError(MindInsightException):
    """Param importance calculated error."""
    def __init__(self, error_msg="Param importance calculated error."):
        super(CorrelationNanError, self).__init__(OptimizerErrors.CORRELATION_NAN,
                                                  error_msg,
                                                  http_code=400)


class HyperConfigError(MindInsightException):
    """Hyper config error."""
    def __init__(self, error_msg="Hyper config is not correct."):
        super(HyperConfigError, self).__init__(OptimizerErrors.HYPER_CONFIG_ERROR,
                                               error_msg,
                                               http_code=400)


class OptimizerTerminateError(MindInsightException):
    """Hyper config error."""
    def __init__(self, error_msg="Auto tuning has been terminated."):
        super(OptimizerTerminateError, self).__init__(OptimizerErrors.OPTIMIZER_TERMINATE,
                                                      error_msg,
                                                      http_code=400)


class ConfigParamError(MindInsightException):
    """Hyper config error."""
    def __init__(self, error_msg="Invalid parameter."):
        super(ConfigParamError, self).__init__(OptimizerErrors.OPTIMIZER_TERMINATE,
                                               error_msg,
                                               http_code=400)


class HyperConfigEnvError(MindInsightException):
    """Hyper config error."""
    def __init__(self, error_msg="Hyper config is not correct."):
        super(HyperConfigEnvError, self).__init__(OptimizerErrors.HYPER_CONFIG_ENV_ERROR,
                                                  error_msg,
                                                  http_code=400)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The memory analyser."""
from enum import Enum
import json
import os

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.exceptions.exceptions import ProfilerIOException, \
    ProfilerFileNotFoundException
from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path
from mindinsight.profiler.common.util import analyse_device_list_from_profiler_dir
from mindinsight.utils.exceptions import ParamValueError


class FileType(Enum):
    """The enum of memory usage file types."""
    SUMMARY = 'summary'
    DETAILS = 'details'


class MemoryUsageAnalyser(BaseAnalyser):
    """Analyse memory usage data from file."""
    _summary_filename = 'memory_usage_summary_{}.json'
    _details_filename = 'memory_usage_details_{}.json'

    def _load(self):
        """Load data according to the parsed profiling files."""

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

    def get_memory_usage_summary(self, device_type):
        """
        Get memory usage summary data for UI display.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.

        Returns:
            json, the content of memory usage summary.
        """
        summary = self._get_file_content(device_type, FileType.SUMMARY.value)
        memory_summary = {'summary': summary}

        return memory_summary

    def get_memory_usage_graphics(self, device_type):
        """
        Get memory usage data for UI display.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.

        Returns:
            json, the content of memory usage data.
        """
        memory_details = self._get_file_content(device_type, FileType.DETAILS.value)
        for graph_id in memory_details.keys():
            if 'breakdowns' in memory_details[graph_id]:
                memory_details[graph_id].pop('breakdowns')

        return memory_details

    def get_memory_usage_breakdowns(self, device_type, graph_id, node_id):
        """
        Get memory usage breakdowns for each node.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.
            graph_id (int): Graph id.
            node_id (int): Node id.

        Returns:
            json, the content of memory usage breakdowns.
        """
        memory_details = self._get_file_content(device_type, FileType.DETAILS.value)
        graph = None
        if memory_details:
            for memory_data in memory_details.values():
                if graph_id == str(memory_data.get('graph_id', -1)):
                    graph = memory_data
                    break
        if not graph:
            logger.error('Invalid graph id: %s', graph_id)
            raise ParamValueError('Invalid graph id.')

        if not ('breakdowns' in graph and node_id < len(graph['breakdowns'])):
            logger.error('Invalid node id: %s', node_id)
            raise ParamValueError('Invalid node id.')

        memory_breakdowns = graph.get('breakdowns')[node_id]

        return {'breakdowns': memory_breakdowns}

    def get_memory_data_for_marey(self, device_type):
        """
        Get memory data for marey's graph.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.

        Returns:
            json, the content of memory usage summary and details for marey's graph.
        """
        data = {}
        device_list, _, _ = analyse_device_list_from_profiler_dir(self._profiling_dir)
        for device in device_list:
            self._device_id = device
            summary = self._get_file_content(device_type, FileType.SUMMARY.value)
            details = self._get_memory_file_without_breakdown(device_type)
            device_entry = "device" + device
            data[device_entry] = {}
            data.get(device_entry)['summary'] = summary
            data.get(device_entry)['details'] = details
        return data

    def _get_file_content(self, device_type, file_type):
        """
        Get file content for different types of memory usage files.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.
            file_type (str): memory usage file type, e.g., summary, details.

        Returns:
            dict, file content corresponding to file_type.
        """
        file_path = self._get_file_path(device_type, file_type)
        if not os.path.exists(file_path):
            logger.error('Invalid file path. Please check the output path: %s', file_path)
            raise ProfilerFileNotFoundException(msg='Invalid memory file path.')

        try:
            with open(file_path, 'r') as f_obj:
                file_content = json.load(f_obj)
        except (IOError, OSError, json.JSONDecodeError) as err:
            logger.error('Error occurred when read memory file: %s', err)
            raise ProfilerIOException()

        return file_content

    def _get_memory_file_without_breakdown(self, device_type):
        """
        Get file content for memory usage details files without breakdowns.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.

        Returns:
            dict, file content corresponding to file_type.
        """
        file_path = self._get_file_path(device_type, FileType.DETAILS.value)
        if not os.path.exists(file_path):
            logger.error('Invalid file path. Please check the output path: %s', file_path)
            raise ProfilerFileNotFoundException(msg='Invalid memory file path.')

        try:
            with open(file_path, 'r') as f_obj:
                f_obj_str = f_obj.read()
                del_start = f_obj_str.find(", \"breakdowns\"")
                new_str = f_obj_str[:del_start] + f_obj_str[-2:]
                del f_obj_str
                file_content = json.loads(new_str)
        except (IOError, OSError, json.JSONDecodeError) as err:
            logger.error('Error occurred when read memory file: %s', err)
            raise ProfilerIOException()

        return file_content

    def _get_file_path(self, device_type, file_type):
        """
        Get memory usage summary file.

        Args:
            device_type (str): Device type, e.g., GPU, Ascend.
            file_type (str): memory usage file type, e.g., summary, details.

        Returns:
            str, file path of memory usage file corresponding to its file_type.
        """
        filename = ""
        if device_type == "ascend":
            if file_type is FileType.SUMMARY.value:
                filename = self._summary_filename.format(self._device_id)
            elif file_type is FileType.DETAILS.value:
                filename = self._details_filename.format(self._device_id)
        else:
            logger.error('Memory Usage only supports Ascend for now. Please check the device type.')
            raise ParamValueError("Invalid device type.")

        file_path = os.path.join(self._profiling_dir, filename)
        file_path = validate_and_normalize_path(
            file_path, raise_key='Invalid memory usage file path.'
        )

        return file_path



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The minddata pipeline analyser class."""
import csv
import json
import os
import sys

from mindinsight.profiler.analyser.base_analyser import BaseAnalyser
from mindinsight.profiler.common.exceptions.exceptions import \
    ProfilerPipelineOpNotExistException
from mindinsight.profiler.common.log import logger
from mindinsight.profiler.common.validator.validate_path import validate_and_normalize_path


class MinddataPipelineAnalyser(BaseAnalyser):
    """
    The analyser for analyzing the minddata pipeline operator and queue data.

    Args:
        profiling_dir (str): The directory where the parsed profiling files are
            located.
        device_id (str): The device ID.

    Raises:
        ProfilerPathErrorException: If the profiling dir is invalid.
    """
    _col_names = ['op_id', 'op_type', 'num_workers', 'output_queue_size',
                  'output_queue_average_size', 'output_queue_length',
                  'output_queue_usage_rate', 'sample_interval', 'parent_id',
                  'children_id']
    _file_name_pipeline = 'minddata_pipeline_raw_{}.csv'
    _index_op_id = 0
    _index_op_type = 1
    _index_num_workers = 2
    _index_output_queue_size = 3
    _index_output_queue_average_size = 4
    _index_output_queue_length = 5
    _index_output_queue_usage_rate = 6
    _index_sample_interval = 7
    _index_parent_id = 8
    _index_children_id = 9

    def __init__(self, profiling_dir, device_id):
        super().__init__(profiling_dir, device_id)
        self._none_filter_condition_key = ['threshold', 'is_display_op_detail']
        self._none_sort_col_names = ['output_queue_size', 'children_id']
        self._op_id_index_map = self._get_op_id_index_map()

    def get_op_and_parent_op_info(self, op_id):
        """
        Get the operator and parent operator information by `op_id`.

        Args:
            op_id (int): The minddata pipeline operator ID.

        Returns:
            dict, the operator and parent operator information.

        Raises:
             ProfilerPipelineOpNotExistException: If the minddata pipeline
                operator does not exist.
        """
        index = self._op_id_index_map.get(op_id)
        if index is None:
            raise ProfilerPipelineOpNotExistException(str(op_id))
        op_info = self._data[index]
        parent_id = op_info[self._index_parent_id]
        parent_index = self._op_id_index_map.get(parent_id)
        if parent_index is None:
            parent_op = None
            queue_info = None
        else:
            parent_op_info = self._data[parent_index]
            parent_op = {
                'op_id': parent_op_info[self._index_op_id],
                'op_type': parent_op_info[self._index_op_type],
                'num_workers': parent_op_info[self._index_num_workers]
            }
            queue_info = {
                'output_queue_size': op_info[self._index_output_queue_size],
                'output_queue_average_size':
                    op_info[self._index_output_queue_average_size],
                'output_queue_length': op_info[self._index_output_queue_length],
                'output_queue_usage_rate':
                    op_info[self._index_output_queue_usage_rate],
                'sample_interval': op_info[self._index_sample_interval]
            }

        current_op = {
            'op_id': op_info[self._index_op_id],
            'op_type': op_info[self._index_op_type],
            'num_workers': op_info[self._index_num_workers]
        }
        return {
            'current_op': current_op,
            'parent_op': parent_op,
            'queue_info': queue_info
        }

    def _load(self):
        """Load data according to the parsed minddata pipeline file."""
        pipeline_file_path = os.path.join(
            self._profiling_dir,
            self._file_name_pipeline.format(self._device_id)
        )
        pipeline_file_path = validate_and_normalize_path(
            pipeline_file_path, raise_key="Invalid pipeline file path.")
        if not os.path.isfile(pipeline_file_path):
            logger.warning('The file <%s> does not exist.', pipeline_file_path)
            return

        with open(pipeline_file_path, 'r') as file:
            csv.field_size_limit(sys.maxsize)
            csv_reader = csv.reader(file)
            _ = next(csv_reader)
            for info in csv_reader:
                self._data.append(self._convert_field_type(info))

    def _filter(self, filter_condition):
        """
        Filter the profiling data according to the filter condition.

        Args:
            filter_condition (dict): The filter condition.
        """

        def _inner_filter(item: list):
            return self._default_filter(item, filter_condition)

        def _inner_map(item: list):
            inner_item = item[0:2]
            inner_item.extend(item[4:])
            return inner_item

        threshold = filter_condition.get('threshold')
        is_display_op_detail = filter_condition.get(
            'is_display_op_detail', False
        )
        self._set_display_col_name(is_display_op_detail)

        filter_result = list(filter(_inner_filter, self._data))
        if threshold:
            low_threshold = threshold[1]
            high_threshold = threshold[0]
            filter_result = self._filter_outside_threshold(
                filter_result, low_threshold, high_threshold
            )

        if is_display_op_detail:
            self._result = filter_result
        else:
            self._result = list(map(_inner_map, filter_result))

    def _filter_outside_threshold(self, data, low_threshold, high_threshold):
        """
        Get the data outside the threshold range.

        Args:
            data (list[list]): The filtered data.
            low_threshold (float): The low threshold.
            high_threshold (float): The high threshold.

        Returns:
            list[list], the data outside the threshold range.
        """
        root_node = None
        leaf_nodes = []
        all_higher_high_threshold = True
        maximum_cur_child_differance = 0.1
        result = []
        for item in data:
            parent_id = item[self._index_parent_id]
            if parent_id is None:
                root_node = item
                continue

            # current usage rate compared to the threshold
            cur_usage_rate = item[self._index_output_queue_usage_rate]
            if cur_usage_rate < high_threshold:
                all_higher_high_threshold = False

            # the child node usage rate compared to the threshold
            child_ids = item[self._index_children_id]
            if not child_ids:
                leaf_nodes.append(item)
                continue
            child_usage_rates = [
                self._get_usage_rate_by_op_id(op_id) for op_id in child_ids
            ]
            for usage_rate in child_usage_rates:
                if usage_rate - cur_usage_rate > maximum_cur_child_differance:
                    if item not in result:
                        result.append(item)

        for leaf_node in leaf_nodes:
            if leaf_node[self._index_output_queue_usage_rate] < low_threshold:
                result.append(leaf_node)

        if all_higher_high_threshold:
            result = [root_node]
        return result

    def _get_usage_rate_by_op_id(self, op_id):
        """
        Gets the usage rate of the queue corresponding to the specified operator.

        Args:
            op_id (int): The pipeline operator ID.

        Returns:
            float, the usage rate of the queue corresponding to the specified
            operator.
        """
        index = self._op_id_index_map.get(op_id)
        op_info = self._data[index]
        return op_info[self._index_output_queue_usage_rate]

    def _set_display_col_name(self, is_display_op_detail):
        """
        Set the display column name according to the filter condition.

        Args:
            is_display_op_detail (bool): Whether to display the detailed operator
                information.
        """
        if not is_display_op_detail:
            self._display_col_names = self._col_names[0:2]
            self._display_col_names.extend(self._col_names[4:])

    def _convert_field_type(self, row):
        """
        Convert the field type of minddata pipeline file to the specific type.

        Args:
            row (list[str]): One row data from parsed data.

        Returns:
            list[Union[str, int, float]], the converted data.
        """
        return [
            int(row[self._index_op_id]),
            row[self._index_op_type],
            int(row[self._index_num_workers]),
            json.loads(row[self._index_output_queue_size])
            if row[self._index_output_queue_size] else None,
            float(row[self._index_output_queue_average_size])
            if row[self._index_output_queue_average_size] else None,
            int(row[self._index_output_queue_length])
            if row[self._index_output_queue_length] else None,
            float(row[self._index_output_queue_usage_rate])
            if row[self._index_output_queue_usage_rate] else 0,
            int(row[self._index_sample_interval]),
            int(row[self._index_parent_id])
            if row[self._index_parent_id] else None,
            json.loads(row[self._index_children_id])
            if row[self._index_children_id] else None
        ]

    def _get_op_id_index_map(self):
        """
        Get the map of the operator id and index in data.

        Returns:
            dict, the map of the operator id and index in data.
        """
        the_map = {}
        for index, op_info in enumerate(self._data):
            the_map[op_info[self._index_op_id]] = index
        return the_map



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Define a validation class which contain all check methods of datavisual module."""
from numbers import Number
from mindinsight.utils.exceptions import ParamValueError
from mindinsight.utils.exceptions import ParamMissError
from mindinsight.datavisual.common.exceptions import PluginNotAvailableError
from mindinsight.datavisual.common.enums import PluginNameEnum
from mindinsight.datavisual.utils.tools import to_int


class Validation:
    """Validation class, define all check methods."""

    @classmethod
    def check_offset(cls, offset, default_value=0):
        """
        Check offset parameter, it must be greater or equal 0.

        Args:
            offset (Union[str, int]): Value can be string number or int.
            default_value (int): Default value for checked offset. Default: 0.

        Returns:
            int, offset.
        """

        if offset is None:
            return default_value
        offset = to_int(offset, 'offset')
        if offset < 0:
            raise ParamValueError("'offset' should be greater than or equal to 0.")
        return offset

    @classmethod
    def check_limit(cls, limit, min_value=1, max_value=1000, default_value=100):
        """
        Check limit parameter, it should between min_value and max_value.

        Args:
            limit (Union[str, int]): Value can be string number or int.
            min_value (int): Limit should greater or equal this value. Default: 1.
            max_value (int): Limit should less or equal this value. Default: 1000.
            default_value (int): Default value for limit. Default: 100.

        Returns:
            int, limit.
        """

        if limit is None:
            return default_value

        limit = to_int(limit, 'limit')
        if limit < min_value or limit > max_value:
            raise ParamValueError("'limit' should in [{}, {}].".format(min_value, max_value))
        return limit

    @classmethod
    def check_param_empty(cls, **kwargs):
        """
        Check param.

        Args:
            kwargs (Any): Check if arg is truthy.

        Raises:
            ParamMissError: When param missing.
        """
        for key, value in kwargs.items():
            # When value is 0, 0.0 or False, it is not empty.
            if isinstance(value, Number):
                continue

            if not value:
                raise ParamMissError(key)

    @classmethod
    def check_plugin_name(cls, plugin_name):
        """
        Check plugin name.

        Args:
            plugin_name (str): The plugin name.

        Raises:
            PluginNotAvailableError: When plugin name is not valid.
        """
        plugin_name_list = PluginNameEnum.list_members()
        if plugin_name not in plugin_name_list:
            raise PluginNotAvailableError(f"'plugin_name' only can be one of {plugin_name_list}")



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Loss landscape container, used for converting proto to python object."""
from mindinsight.datavisual.data_transform.tensor_container import TensorContainer


class PointContainer:
    """
    Point data container.

    Args:
        point (mindinsight_summary.proto.LossLandscape.Point): landscape point data.
    """
    def __init__(self, point):
        self._x = TensorContainer(point.x)
        self._y = TensorContainer(point.y)
        self._z = TensorContainer(point.z)

    @property
    def x(self):
        """Get the x-coordinate of the point."""
        return self._x

    @property
    def y(self):
        """Get the y-coordinate of the point."""
        return self._y

    @property
    def z(self):
        """Get the z-coordinate of the point."""
        return self._z


class LossPathContainer:
    """
    LossPath data container.

    Args:
        loss_point (mindinsight_summary.proto.LossLandscape.LossPath): loss path data.
    """
    def __init__(self, loss_point):
        self._intervals = [interval for interval in loss_point.intervals]
        self._points = PointContainer(loss_point.points)

    @property
    def intervals(self):
        """Get the intervals."""
        return self._intervals

    @property
    def points(self):
        """Get the points."""
        return self._points


class MetadataContainer:
    """Metadata container."""
    def __init__(self, metadata):
        self._decomposition = metadata.decomposition
        self._unit = metadata.unit
        self._step_per_epoch = metadata.step_per_epoch

    @property
    def decomposition(self):
        """Get the decomposition."""
        return self._decomposition

    @property
    def unit(self):
        """Get the unit."""
        return self._unit

    @property
    def step_per_epoch(self):
        """Get the step per epoch."""
        return self._step_per_epoch


class LossLandscapeContainer:
    """Loss landscape container."""
    def __init__(self, loss_landscape_message):
        self._landscape = PointContainer(loss_landscape_message.landscape)
        self._loss_path = LossPathContainer(loss_landscape_message.loss_path)
        self._metadata = MetadataContainer(loss_landscape_message.metadata)
        self._convergence_point = PointContainer(loss_landscape_message.convergence_point)

    @property
    def landscape(self):
        """Get the landscape."""
        return self._landscape

    @property
    def loss_path(self):
        """Get the loss_path."""
        return self._loss_path

    @property
    def metadata(self):
        """Get the metadata."""
        return self._metadata

    @property
    def convergence_point(self):
        """Get the convergence_point."""
        return self._convergence_point



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Web application module."""
import os
from importlib import import_module
from copy import deepcopy
from werkzeug.datastructures import Headers
from werkzeug.exceptions import HTTPException

from flask import Flask
from flask import request
from flask import Response

from mindinsight.conf import settings
from mindinsight.utils.hook import HookUtils
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.common.exceptions import RequestMethodNotAllowed
from mindinsight.datavisual.common import error_handler
from mindinsight.datavisual.utils.tools import find_app_package
from mindinsight.datavisual.utils.tools import get_img_mimetype
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.utils.log import setup_logger


def get_security_headers():
    """Get security headers."""
    domain_white_list = []
    for hook in HookUtils.instance().hooks():
        domain_white_list += hook.register_secure_domains()

    content_security_policy = {
        'img-src': ["'self'", 'data:', 'blob:'],
        'style-src': ["'self'", "'unsafe-inline'"],
        'frame-src': ["'self'"] + domain_white_list,
        'frame-ancestors': ["'self'"] + domain_white_list,
        'default-src': ["'self'"],
        'script-src': ["'self'", "'unsafe-eval'"]
    }

    request_methods = deepcopy(settings.SUPPORT_REQUEST_METHODS)
    if settings.ENABLE_CORS:
        request_methods.add('OPTIONS')

    headers = {
        'X-Frame-Options': 'SAMEORIGIN',
        'X-XSS-Protection': '1; mode=block',
        'X-Content-Type-Options': 'nosniff',
        'Access-Control-Allow-Methods': ', '.join(request_methods),
        'Content-Security-Policy': '; '.join([
            f"{k} {' '.join(v)}" for k, v in content_security_policy.items()
        ]),
        'X-Download-Options': 'noopen',
        'Cache-Control': 'no-store',
        'Pragma': 'no-cache'
    }

    return list(headers.items())


SECURITY_HEADERS = get_security_headers()


class CustomResponse(Response):
    """Define custom response."""
    def __init__(self, response=None, **kwargs):
        headers = kwargs.get("headers")
        security_headers = list(SECURITY_HEADERS)
        if isinstance(response, bytes):
            mimetype = get_img_mimetype(response)
            security_headers.append(('Content-Type', mimetype))
        if headers is None:
            headers = Headers(security_headers)
        else:
            for header in security_headers:
                headers.add(*header)
        kwargs['headers'] = headers
        super(CustomResponse, self).__init__(response, **kwargs)


def _init_app_module(app):
    """
    Init app module.

    Args:
        app (Flask): An instance of Flask.
    """
    packages = find_app_package()
    gunicorn_logger = setup_logger("gunicorn", "error")
    for package in packages:
        try:
            app_module = import_module(package)
            gunicorn_logger.info("[%s].init_module starts.", package)
            app_module.init_module(app)
            gunicorn_logger.info("[%s].init_module ends.", package)
        except AttributeError:
            logger.debug('[%s].init_module not exists.', package)


def before_request():
    """A function to run before each request."""
    if request.method not in settings.SUPPORT_REQUEST_METHODS:
        raise RequestMethodNotAllowed()


def create_app():
    """Set flask APP config, and start the data manager."""
    gunicorn_logger = setup_logger("gunicorn", "error")
    gunicorn_logger.info("create_app starts.")
    static_url_path = settings.URL_PATH_PREFIX + "/static"
    static_folder_path = os.path.realpath(os.path.join(os.path.dirname(__file__), os.pardir, 'ui', 'dist', 'static'))

    app = Flask(__name__, static_url_path=static_url_path, static_folder=static_folder_path)
    app.config['JSON_SORT_KEYS'] = False

    if settings.ENABLE_CORS:
        flask_cors = import_module('flask_cors')
        flask_cors.CORS(app, supports_credentials=True)

    app.before_request(before_request)

    app.register_error_handler(HTTPException, error_handler.handle_http_exception_error)
    app.register_error_handler(MindInsightException, error_handler.handle_mindinsight_error)
    app.register_error_handler(Exception, error_handler.handle_unknown_error)

    app.response_class = CustomResponse

    _init_app_module(app)
    gunicorn_logger.info("create_app ends.")

    return app


APP = create_app()



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""DebuggerTensor."""
import re
from abc import ABC

import numpy as np

from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.utils import NUMPY_TYPE_MAP
from mindinsight.debugger.stream_cache.data_loader import DumpTarget
from mindinsight.domain.graph.base import NodeType
from mindinsight.domain.graph.proto.ms_graph_pb2 import DataType


class DebuggerTensor(ABC):
    """
    The tensor with specific rank, iteration and debugging info.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        node (Node): The node that outputs this tensor.
        slot (int): The slot of the tensor on the node.
        iteration (int): The iteration of the tensor.

    Note:
        - Users should not instantiate this class manually.
        - The instances of this class is immutable.
        - A `DebuggerTensor` is always the output tensor of a node.
    """

    def __init__(self, node, slot, iteration):
        self._node = node
        self._slot = slot
        self._iteration = iteration

    @property
    def node(self):
        """
        Get the node that outputs this tensor.

        Returns:
            Node, the node that outputs this tensor.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> tensors = list(my_run.select_tensors("conv"))
                >>> print(tensors[0].node)
                rank: 0
                graph_name: kernel_graph_0
                node_name: conv1.weight
        """
        return self._node

    @property
    def slot(self):
        """
        The output of the node may have several tensors. The slot refer to the index of the tensor

        Returns:
            int, the slot of the tensor on the node.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> tensors = list(my_run.select_tensors("conv"))
                >>> print(tensors[0].slot)
                0
        """
        return self._slot

    @property
    def iteration(self):
        """
        Get iteration of the tensor.

        Returns:
            int, the iteration of the tensor.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> tensors = list(my_run.select_tensors("conv"))
                >>> print(tensors[0].iteration)
                0
        """
        return self._iteration

    @property
    def rank(self):
        """
        The rank is the logical id of the device on which the tensor is generated.

        Returns:
            int, the rank for this tensor.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> tensors = list(my_run.select_tensors("conv"))
                >>> print(tensors[0].rank)
                0
        """
        return self._node.rank

    def value(self):
        """
        Get the value of the tensor.

        Returns:
            Union[numpy.array, None], The value could be None if failed to find data file
            in relative iteration.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>>
                >>> def test_debugger_tensor():
                ...     my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                ...     tensors = list(my_run.select_tensors("conv"))
                ...     # the tensors[0].value() maybe start the new process
                ...     value = tensors[0].value()
                ...     return value
                ...
                >>> if __name__ == "__main__":
                ...     test_debugger_tensor()
                ...
        """
        raise NotImplementedError

    def __str__(self):
        feature = f"rank: {self.rank}\n" \
                  f"graph_name: {self.node.graph_name}\n" \
                  f"node_name: {self.node.name}\n" \
                  f"slot: {self.slot}\n" \
                  f"iteration: {self.iteration}"
        return feature


class DebuggerTensorImpl(DebuggerTensor):
    """DebuggerTensor implementation."""

    @property
    def root_graph_id(self):
        """Get the root_graph_id for this tensor."""
        return self._node.root_graph_id

    def has_value(self):
        """Check if the tensor has value."""
        iteration = self.iteration
        if iteration is None:
            return False
        data_loader = self.node.debugger_engine.data_loader
        has_dump_output = bool(data_loader.dump_target in [DumpTarget.FULL, DumpTarget.OUTPUT_ONLY])
        if not has_dump_output:
            return False
        if self.node.node_type == NodeType.CONSTANT:
            iteration = 'Constant'
        iter_dirs = data_loader.get_step_iter(rank_id=self.rank, step=iteration)
        file_found = self._file_found(iter_dirs)
        return file_found

    def _file_found(self, iter_dirs):
        """Check if the tensor file found in specified directory."""
        node_name_without_scope = self.node.name.split('/')[-1]
        bin_pattern = node_name_without_scope + r".*.(\d+)$"
        npy_pattern = f"{node_name_without_scope}.*.output.{self.slot}.*.npy$"
        for iter_dir in iter_dirs:
            for tensor_path in iter_dir.iterdir():
                file_name = tensor_path.name
                if re.search(bin_pattern, file_name) or re.search(npy_pattern, file_name):
                    return True
        return False

    def value(self):
        if self.iteration is None:
            log.warning("The iteration of is not specified, no value returned.")
            return None
        base_node = self.node.base_node
        if hasattr(base_node, 'output') and hasattr(base_node.output, 'info'):
            info = base_node.output.info
            if isinstance(info, dict) and info.get("np_value") is not None:
                return info.get("np_value")
        debugger_engine = self.node.debugger_engine
        tensor_info = debugger_engine.dbg_services_module.TensorInfo(
            node_name=base_node.full_name if self.node.node_type == NodeType.CONSTANT else self.node.name,
            slot=self.slot,
            iteration=self.iteration,
            rank_id=self.rank,
            root_graph_id=self.root_graph_id,
            is_output=True)
        tensors = debugger_engine.dbg_service.read_tensors([tensor_info])
        return self._to_numpy(tensors[0])

    @staticmethod
    def _to_numpy(tensor_data):
        """Turn tensor data into Numpy."""
        if tensor_data.data_size == 0:
            return None
        dtype_str = DataType.Name(tensor_data.dtype)
        np_type = NUMPY_TYPE_MAP.get(dtype_str)
        data = np.frombuffer(tensor_data.data_ptr, dtype=np_type)
        data = data.reshape(tensor_data.shape)
        return data



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Histogram data."""
import math

from mindinsight.utils.exceptions import ParamValueError
from mindinsight.datavisual.utils.utils import calc_histogram_bins


def mask_invalid_number(num):
    """Mask invalid number to 0."""
    if math.isnan(num) or math.isinf(num):
        return type(num)(0)

    return num


class Bucket:
    """
    Bucket data class.

    Args:
        left (double): Left edge of the histogram bucket.
        width (double): Width of the histogram bucket.
        count (int): Count of numbers fallen in the histogram bucket.
    """
    def __init__(self, left, width, count):
        self._left = left
        self._width = width
        self._count = count

    @property
    def left(self):
        """Gets left edge of the histogram bucket."""
        return self._left

    @property
    def count(self):
        """Gets count of numbers fallen in the histogram bucket."""
        return self._count

    @property
    def width(self):
        """Gets width of the histogram bucket."""
        return self._width

    @property
    def right(self):
        """Gets right edge of the histogram bucket."""
        return self._left + self._width

    def as_tuple(self):
        """Gets the bucket as tuple."""
        return self._left, self._width, self._count

    def __repr__(self):
        """Returns repr(self)."""
        return "Bucket(left={}, width={}, count={})".format(self._left, self._width, self._count)


class Histogram:
    """
    Histogram data class.

    Args:
        buckets (tuple[Bucket]): The buckets of histogram data.
        max_val (number): The max value of histogram data.
        min_val (number): The min value of histogram data.
        count (int): The count of histogram data.
    """

    # Max quantity of original buckets.
    MAX_ORIGINAL_BUCKETS_COUNT = 90

    def __init__(self, buckets, max_val, min_val, count):
        self._visual_max = max_val
        self._visual_min = min_val
        self._count = count
        self._original_buckets = buckets
        # default bin number
        self._visual_bins = calc_histogram_bins(count)
        # Note that tuple is immutable, so sharing tuple is often safe.
        self._re_sampled_buckets = ()

    @property
    def original_buckets_count(self):
        """Gets original buckets quantity."""
        return len(self._original_buckets)

    def set_visual_range(self, max_val: float, min_val: float, bins: int) -> None:
        """
        Sets visual range for later re-sampling.

        It's caller's duty to ensure input is valid.

        Why we need visual range for histograms? Aligned buckets between steps can help users know about the trend of
        tensors. Miss aligned buckets between steps might miss-lead users about the trend of a tensor. Because for
        given tensor, if you have thinner buckets, count of every bucket will get lower, however, if you have
        thicker buckets, count of every bucket will get higher.  When they are displayed together, user might think
        the histogram with thicker buckets has more values. This is miss-leading. So we need to unify buckets across
        steps. Visual range for histogram is a technology for unifying buckets.

        Args:
            max_val (float): Max value for visual histogram.
            min_val (float): Min value for visual histogram.
            bins (int): Bins number for visual histogram.
        """
        if max_val < min_val:
            raise ParamValueError(
                "Invalid input. max_val({}) is less or equal than min_val({}).".format(max_val, min_val))

        if bins < 1:
            raise ParamValueError("Invalid input bins({}). Must be greater than 0.".format(bins))

        self._visual_max = max_val
        self._visual_min = min_val
        self._visual_bins = bins

        # mark _re_sampled_buckets to empty
        self._re_sampled_buckets = ()

    def _calc_intersection_len(self, max1, min1, max2, min2):
        """Calculates intersection length of [min1, max1] and [min2, max2]."""
        if max1 < min1:
            raise ParamValueError(
                "Invalid input. max1({}) is less than min1({}).".format(max1, min1))

        if max2 < min2:
            raise ParamValueError(
                "Invalid input. max2({}) is less than min2({}).".format(max2, min2))

        if min1 <= min2:
            if max1 <= min2:
                # return value must be calculated by max1.__sub__
                return max1 - max1
            if max1 <= max2:
                return max1 - min2
            # max1 > max2
            return max2 - min2

        # min1 > min2
        if max2 <= min1:
            return max2 - max2
        if max2 <= max1:
            return max2 - min1
        return max1 - min1

    def _re_sample_buckets(self):
        """Re-samples buckets according to visual_max, visual_min and visual_bins."""
        if self._visual_max == self._visual_min:
            # Adjust visual range if max equals min.
            self._visual_max += 0.5
            self._visual_min -= 0.5

        width = (self._visual_max - self._visual_min) / self._visual_bins

        if not self._count:
            self._re_sampled_buckets = tuple(
                Bucket(self._visual_min + width * i, width, 0)
                for i in range(self._visual_bins))
            return

        re_sampled = []
        original_pos = 0
        original_bucket = self._original_buckets[original_pos]
        for i in range(self._visual_bins):
            cur_left = self._visual_min + width * i
            cur_right = cur_left + width
            cur_estimated_count = 0.0

            # Skip no bucket range.
            if cur_right <= original_bucket.left:
                re_sampled.append(Bucket(cur_left, width, math.ceil(cur_estimated_count)))
                continue

            # Skip no intersect range.
            while cur_left >= original_bucket.right:
                original_pos += 1
                if original_pos >= len(self._original_buckets):
                    break
                original_bucket = self._original_buckets[original_pos]

            # entering with this condition: cur_right > original_bucket.left and cur_left < original_bucket.right
            while True:
                if original_pos >= len(self._original_buckets):
                    break
                original_bucket = self._original_buckets[original_pos]

                intersection = self._calc_intersection_len(
                    min1=cur_left, max1=cur_right,
                    min2=original_bucket.left, max2=original_bucket.right)
                if not original_bucket.width:
                    estimated_count = original_bucket.count
                else:
                    estimated_count = (intersection / original_bucket.width) * original_bucket.count

                cur_estimated_count += estimated_count
                if cur_right > original_bucket.right:
                    # Need to sample next original bucket to this visual bucket.
                    original_pos += 1
                else:
                    # Current visual bucket has taken all intersect buckets into account.
                    break

            re_sampled.append(Bucket(cur_left, width, math.ceil(cur_estimated_count)))

        self._re_sampled_buckets = tuple(re_sampled)

    def buckets(self, convert_to_tuple=True):
        """
        Get visual buckets instead of original buckets.

        Args:
            convert_to_tuple (bool): Whether convert bucket object to tuple.

        Returns:
            tuple, contains buckets.
        """
        if not self._re_sampled_buckets:
            self._re_sample_buckets()

        if not convert_to_tuple:
            return self._re_sampled_buckets

        return tuple(bucket.as_tuple() for bucket in self._re_sampled_buckets)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mindinsight_xai.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='mindinsight_xai.proto',
  package='xai',
  syntax='proto2',
  serialized_options=_b('\370\001\001'),
  serialized_pb=_b('\n\x15mindinsight_xai.proto\x12\x03xai\"d\n\x05\x45vent\x12\x11\n\twall_time\x18\x01 \x02(\x01\x12\x0c\n\x04step\x18\x02 \x01(\x03\x12\x11\n\x07version\x18\x03 \x01(\tH\x00\x12\x1f\n\x07\x65xplain\x18\x06 \x01(\x0b\x32\x0c.xai.ExplainH\x00\x42\x06\n\x04what\"\xbb\x07\n\x07\x45xplain\x12\x11\n\tsample_id\x18\x01 \x01(\x05\x12\x12\n\nimage_path\x18\x02 \x01(\t\x12\x1a\n\x12ground_truth_label\x18\x03 \x03(\x05\x12)\n\tinference\x18\x04 \x01(\x0b\x32\x16.xai.Explain.Inference\x12-\n\x0b\x65xplanation\x18\x05 \x03(\x0b\x32\x18.xai.Explain.Explanation\x12)\n\tbenchmark\x18\x06 \x03(\x0b\x32\x16.xai.Explain.Benchmark\x12\'\n\x08metadata\x18\x07 \x01(\x0b\x32\x15.xai.Explain.Metadata\x12\x0e\n\x06status\x18\x08 \x01(\t\x12\x1d\n\x03hoc\x18\t \x03(\x0b\x32\x10.xai.Explain.Hoc\x1a\x9c\x02\n\tInference\x12\x19\n\x11ground_truth_prob\x18\x01 \x03(\x02\x12\x17\n\x0fpredicted_label\x18\x02 \x03(\x05\x12\x16\n\x0epredicted_prob\x18\x03 \x03(\x02\x12\x1c\n\x14ground_truth_prob_sd\x18\x04 \x03(\x02\x12#\n\x1bground_truth_prob_itl95_low\x18\x05 \x03(\x02\x12\"\n\x1aground_truth_prob_itl95_hi\x18\x06 \x03(\x02\x12\x19\n\x11predicted_prob_sd\x18\x07 \x03(\x02\x12 \n\x18predicted_prob_itl95_low\x18\x08 \x03(\x02\x12\x1f\n\x17predicted_prob_itl95_hi\x18\t \x03(\x02\x1aJ\n\x0b\x45xplanation\x12\x16\n\x0e\x65xplain_method\x18\x01 \x01(\t\x12\r\n\x05label\x18\x02 \x01(\x05\x12\x14\n\x0cheatmap_path\x18\x03 \x01(\t\x1ag\n\tBenchmark\x12\x18\n\x10\x62\x65nchmark_method\x18\x01 \x01(\t\x12\x16\n\x0e\x65xplain_method\x18\x02 \x01(\t\x12\x13\n\x0btotal_score\x18\x03 \x01(\x02\x12\x13\n\x0blabel_score\x18\x04 \x03(\x02\x1aK\n\x08Metadata\x12\r\n\x05label\x18\x01 \x03(\t\x12\x16\n\x0e\x65xplain_method\x18\x02 \x03(\t\x12\x18\n\x10\x62\x65nchmark_method\x18\x03 \x03(\t\x1a%\n\x08HocLayer\x12\x0c\n\x04prob\x18\x01 \x01(\x02\x12\x0b\n\x03\x62ox\x18\x02 \x03(\x05\x1aH\n\x03Hoc\x12\r\n\x05label\x18\x01 \x01(\x05\x12\x0c\n\x04mask\x18\x02 \x01(\t\x12$\n\x05layer\x18\x03 \x03(\x0b\x32\x15.xai.Explain.HocLayerB\x03\xf8\x01\x01')
)




_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='xai.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='wall_time', full_name='xai.Event.wall_time', index=0,
      number=1, type=1, cpp_type=5, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='step', full_name='xai.Event.step', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='xai.Event.version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='explain', full_name='xai.Event.explain', index=3,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='what', full_name='xai.Event.what',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=30,
  serialized_end=130,
)


_EXPLAIN_INFERENCE = _descriptor.Descriptor(
  name='Inference',
  full_name='xai.Explain.Inference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ground_truth_prob', full_name='xai.Explain.Inference.ground_truth_prob', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicted_label', full_name='xai.Explain.Inference.predicted_label', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicted_prob', full_name='xai.Explain.Inference.predicted_prob', index=2,
      number=3, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ground_truth_prob_sd', full_name='xai.Explain.Inference.ground_truth_prob_sd', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ground_truth_prob_itl95_low', full_name='xai.Explain.Inference.ground_truth_prob_itl95_low', index=4,
      number=5, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ground_truth_prob_itl95_hi', full_name='xai.Explain.Inference.ground_truth_prob_itl95_hi', index=5,
      number=6, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicted_prob_sd', full_name='xai.Explain.Inference.predicted_prob_sd', index=6,
      number=7, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicted_prob_itl95_low', full_name='xai.Explain.Inference.predicted_prob_itl95_low', index=7,
      number=8, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='predicted_prob_itl95_hi', full_name='xai.Explain.Inference.predicted_prob_itl95_hi', index=8,
      number=9, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=433,
  serialized_end=717,
)

_EXPLAIN_EXPLANATION = _descriptor.Descriptor(
  name='Explanation',
  full_name='xai.Explain.Explanation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='explain_method', full_name='xai.Explain.Explanation.explain_method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label', full_name='xai.Explain.Explanation.label', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='heatmap_path', full_name='xai.Explain.Explanation.heatmap_path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=719,
  serialized_end=793,
)

_EXPLAIN_BENCHMARK = _descriptor.Descriptor(
  name='Benchmark',
  full_name='xai.Explain.Benchmark',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='benchmark_method', full_name='xai.Explain.Benchmark.benchmark_method', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='explain_method', full_name='xai.Explain.Benchmark.explain_method', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_score', full_name='xai.Explain.Benchmark.total_score', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label_score', full_name='xai.Explain.Benchmark.label_score', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=795,
  serialized_end=898,
)

_EXPLAIN_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='xai.Explain.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='xai.Explain.Metadata.label', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='explain_method', full_name='xai.Explain.Metadata.explain_method', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='benchmark_method', full_name='xai.Explain.Metadata.benchmark_method', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=900,
  serialized_end=975,
)

_EXPLAIN_HOCLAYER = _descriptor.Descriptor(
  name='HocLayer',
  full_name='xai.Explain.HocLayer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='prob', full_name='xai.Explain.HocLayer.prob', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='box', full_name='xai.Explain.HocLayer.box', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=977,
  serialized_end=1014,
)

_EXPLAIN_HOC = _descriptor.Descriptor(
  name='Hoc',
  full_name='xai.Explain.Hoc',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='xai.Explain.Hoc.label', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mask', full_name='xai.Explain.Hoc.mask', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='layer', full_name='xai.Explain.Hoc.layer', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1016,
  serialized_end=1088,
)

_EXPLAIN = _descriptor.Descriptor(
  name='Explain',
  full_name='xai.Explain',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sample_id', full_name='xai.Explain.sample_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='image_path', full_name='xai.Explain.image_path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ground_truth_label', full_name='xai.Explain.ground_truth_label', index=2,
      number=3, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inference', full_name='xai.Explain.inference', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='explanation', full_name='xai.Explain.explanation', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='benchmark', full_name='xai.Explain.benchmark', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='xai.Explain.metadata', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='status', full_name='xai.Explain.status', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hoc', full_name='xai.Explain.hoc', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_EXPLAIN_INFERENCE, _EXPLAIN_EXPLANATION, _EXPLAIN_BENCHMARK, _EXPLAIN_METADATA, _EXPLAIN_HOCLAYER, _EXPLAIN_HOC, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=133,
  serialized_end=1088,
)

_EVENT.fields_by_name['explain'].message_type = _EXPLAIN
_EVENT.oneofs_by_name['what'].fields.append(
  _EVENT.fields_by_name['version'])
_EVENT.fields_by_name['version'].containing_oneof = _EVENT.oneofs_by_name['what']
_EVENT.oneofs_by_name['what'].fields.append(
  _EVENT.fields_by_name['explain'])
_EVENT.fields_by_name['explain'].containing_oneof = _EVENT.oneofs_by_name['what']
_EXPLAIN_INFERENCE.containing_type = _EXPLAIN
_EXPLAIN_EXPLANATION.containing_type = _EXPLAIN
_EXPLAIN_BENCHMARK.containing_type = _EXPLAIN
_EXPLAIN_METADATA.containing_type = _EXPLAIN
_EXPLAIN_HOCLAYER.containing_type = _EXPLAIN
_EXPLAIN_HOC.fields_by_name['layer'].message_type = _EXPLAIN_HOCLAYER
_EXPLAIN_HOC.containing_type = _EXPLAIN
_EXPLAIN.fields_by_name['inference'].message_type = _EXPLAIN_INFERENCE
_EXPLAIN.fields_by_name['explanation'].message_type = _EXPLAIN_EXPLANATION
_EXPLAIN.fields_by_name['benchmark'].message_type = _EXPLAIN_BENCHMARK
_EXPLAIN.fields_by_name['metadata'].message_type = _EXPLAIN_METADATA
_EXPLAIN.fields_by_name['hoc'].message_type = _EXPLAIN_HOC
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['Explain'] = _EXPLAIN
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'mindinsight_xai_pb2'
  # @@protoc_insertion_point(class_scope:xai.Event)
  })
_sym_db.RegisterMessage(Event)

Explain = _reflection.GeneratedProtocolMessageType('Explain', (_message.Message,), {

  'Inference' : _reflection.GeneratedProtocolMessageType('Inference', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAIN_INFERENCE,
    '__module__' : 'mindinsight_xai_pb2'
    # @@protoc_insertion_point(class_scope:xai.Explain.Inference)
    })
  ,

  'Explanation' : _reflection.GeneratedProtocolMessageType('Explanation', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAIN_EXPLANATION,
    '__module__' : 'mindinsight_xai_pb2'
    # @@protoc_insertion_point(class_scope:xai.Explain.Explanation)
    })
  ,

  'Benchmark' : _reflection.GeneratedProtocolMessageType('Benchmark', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAIN_BENCHMARK,
    '__module__' : 'mindinsight_xai_pb2'
    # @@protoc_insertion_point(class_scope:xai.Explain.Benchmark)
    })
  ,

  'Metadata' : _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAIN_METADATA,
    '__module__' : 'mindinsight_xai_pb2'
    # @@protoc_insertion_point(class_scope:xai.Explain.Metadata)
    })
  ,

  'HocLayer' : _reflection.GeneratedProtocolMessageType('HocLayer', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAIN_HOCLAYER,
    '__module__' : 'mindinsight_xai_pb2'
    # @@protoc_insertion_point(class_scope:xai.Explain.HocLayer)
    })
  ,

  'Hoc' : _reflection.GeneratedProtocolMessageType('Hoc', (_message.Message,), {
    'DESCRIPTOR' : _EXPLAIN_HOC,
    '__module__' : 'mindinsight_xai_pb2'
    # @@protoc_insertion_point(class_scope:xai.Explain.Hoc)
    })
  ,
  'DESCRIPTOR' : _EXPLAIN,
  '__module__' : 'mindinsight_xai_pb2'
  # @@protoc_insertion_point(class_scope:xai.Explain)
  })
_sym_db.RegisterMessage(Explain)
_sym_db.RegisterMessage(Explain.Inference)
_sym_db.RegisterMessage(Explain.Explanation)
_sym_db.RegisterMessage(Explain.Benchmark)
_sym_db.RegisterMessage(Explain.Metadata)
_sym_db.RegisterMessage(Explain.HocLayer)
_sym_db.RegisterMessage(Explain.Hoc)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
This file is to process `data_transform.data_manager` to handle graph,
and the status of graph will be checked before calling `Graph` object.
"""

from mindinsight.datavisual.common import exceptions
from mindinsight.datavisual.common.enums import PluginNameEnum
from mindinsight.datavisual.common.validation import Validation
from mindinsight.datavisual.processors.base_processor import BaseProcessor
from mindinsight.datavisual.common.exceptions import NodeNotInGraphError


class GraphProcessor(BaseProcessor):
    """
    This object is to handle `DataManager` object, and process graph object.

    Args:
        train_id (str): To get train job data by this given id.
        data_manager (DataManager): A `DataManager` object.
        tag (str): The tag of graph, if tag is None, will load the first graph.
    """
    def __init__(self, train_id, data_manager, tag=None, mode='normal'):
        Validation.check_param_empty(train_id=train_id)
        super(GraphProcessor, self).__init__(data_manager)
        plugin = PluginNameEnum.GRAPH.value if mode == 'normal' else PluginNameEnum.OPTIMIZED_GRAPH.value
        train_job = self._data_manager.get_train_job_by_plugin(train_id, plugin)
        if train_job is None:
            raise exceptions.TrainJobNotExistError()
        if not train_job['tags'] or (tag is not None and tag not in train_job['tags']):
            raise exceptions.GraphNotExistError()

        if tag is None:
            tag = train_job['tags'][0]

        tensors = self._data_manager.list_tensors(train_id, tag=tag)
        self._graph = tensors[0].value

    def list_nodes(self, scope):
        """
        Get the nodes of every layer in graph.

        Args:
            scope (str): The name of a scope.

        Returns:
            TypedDict('Nodes', {'nodes': list[Node]}), format is {'nodes': [<Node object>]}.
                example:
                    {
                      "nodes" : [
                        {
                          "attr" :
                          {
                            "index" : "i: 0\n"
                          },
                          "input" : {},
                          "name" : "input_tensor",
                          "output" :
                          {
                            "Default/TensorAdd-op17" :
                            {
                              "edge_type" : "data",
                              "scope" : "name_scope",
                              "shape" : [1, 16, 128, 128]
                            }
                          },
                          "output_i" : -1,
                          "proxy_input" : {},
                          "proxy_output" : {},
                          "independent_layout" : False,
                          "subnode_count" : 0,
                          "type" : "Data"
                        }
                      ]
                    }
        """
        if scope and not self._graph.exist_node(scope):
            raise NodeNotInGraphError(node_name=scope)

        nodes = self._graph.list_node_by_scope(scope=scope)
        return {'nodes': nodes}

    def search_node_names(self, search_content, offset, limit):
        """
        Search node names by search content.

        Args:
            search_content (Any): This content can be the key content of the node to search.
            offset (int): An offset for page. Ex, offset is 0, mean current page is 1.
            limit (int): The max data items for per page.

        Returns:
            Dict, the searched nodes.
        """
        offset = Validation.check_offset(offset=offset)
        limit = Validation.check_limit(limit, min_value=1, max_value=1000)
        nodes = self._graph.search_nodes_by_pattern(search_content)
        real_offset = offset * limit
        search_nodes = self._graph.get_nodes(nodes[real_offset:real_offset + limit])

        return {"nodes": search_nodes}

    def search_single_node(self, name):
        """
        Search node by node name.

        Args:
            name (str): The name of node.

        Returns:
            dict, format is:
                item_object = {'nodes': [<Node object>],
                       'scope_name': '',
                       'children': {<item_object>}}
        """
        Validation.check_param_empty(name=name)

        nodes = self._graph.search_single_node(name)
        return nodes

    def list_all_nodes_info(self, scope):
        """
        Get all the nodes info in graph.

        Args:
            scope (str): The name of a scope.

        Returns:
            TypedDict('Nodes', {'nodes': list[Node]}), format is {'nodes': [<Node object>]}.
                example:
                    {
                      "nodes" : [
                        {
                          "attr" :
                          {
                            "index" : "i: 0\n"
                          },
                          "input" : {},
                          "name" : "input_tensor",
                          "output" :
                          {
                            "Default/TensorAdd-op17" :
                            {
                              "edge_type" : "data",
                              "scope" : "name_scope",
                              "shape" : [1, 16, 128, 128]
                            }
                          },
                          "independent_layout" : False,
                          "subnode_count" : 0,
                          "type" : "Data"
                        }
                      ]
                    }
        """
        all_nodes_detail = []
        if scope and not self._graph.exist_node(scope):
            raise NodeNotInGraphError(node_name=scope)

        all_nodes_detail = self.get_all_node_detail(scope, all_nodes_detail)
        return {"all_nodes_detail": all_nodes_detail}

    def get_all_node_detail(self, scope, all_nodes_detail):
        """
        get all the nodes detail information recursively.

        Args:
            scope (str): The name of node.
            all_nodes_detail (list): all node info

        Returns:
            list, format is:
                item_object = [{'nodes': [<Node object>],
                       'scope_name': '',
                       'children': {<item_object>}}]
        """
        if scope and not self._graph.exist_node(scope):
            raise NodeNotInGraphError(node_name=scope)

        nodes = self._graph.list_node_by_scope(scope=scope)
        if not isinstance(nodes, list):
            raise NodeNotInGraphError(node_name=scope)
        if not nodes:
            return all_nodes_detail
        for node in nodes:
            name = node.get("name")
            sub_scope = node.get("subnode_count")
            if sub_scope == 0:
                all_nodes_detail.append(node)
                continue
            self.get_all_node_detail(name, all_nodes_detail)
        return all_nodes_detail



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The cluster parallel proposer."""
import os

from mindinsight.profiler.proposer.allproposers.base_proposer import Proposer
from mindinsight.profiler.analyser.analyser_factory import AnalyserFactory
from mindinsight.profiler.common.log import logger


class ParallelProposer(Proposer):
    """The cluster parallel proposer."""

    _DATA_PARALLEL = 'data-parallel'
    _MODEL_PARALLEL = 'model-parallel'
    _PIPELINE_PARALLEL = 'pipeline-parallel'

    # the threshold for cluster data
    _step_interval_threshold = 0.2
    _flops_per_second_threshold = 0.2
    _rank_link_threshold = 0.2
    _data_parallel_communication_threshold = 0.1
    _model_parallel_communication_threshold = 0.1
    _between_stage_communication_threshold = 0.1
    _within_stage_communication_threshold = 0.1
    _flops_threshold = 0.2

    _VAL_DECIMAL_PLACES = 4
    _PERCENT_DECIMAL_PLACES = 2

    def __init__(self, profiling_dir, rank_id):
        super().__init__(profiling_dir, rank_id)
        self._parallel_analyser = AnalyserFactory.instance().get_analyser(
            'parallel', self.profiling_path, self.rank_id)
        self._parallel_mode = self._parallel_analyser.parallel_mode
        self._stage_num = self._parallel_analyser.stage_num
        self._rank_size = self._parallel_analyser.rank_size

        self._proposal_dict = {
            "step_interval": {},
            "flops_per_second": {},
            "cluster_link": [],
            "parallel_strategy": {
                "data_parallel_communication_rate": {},
                "model_parallel_communication_rate": {},
                "pipeline_parallel_between_stage": {},
                "pipeline_parallel_within_stage": {},
                "pipeline_parallel_flops": {}
            }
        }

        self._analyse_hccl = True
        self._analyse_flops = True
        self._analyse_op_segmentation = True
        self._validate_parallel_profiler_files()

    def analyze(self, options=None):
        """
        Get the proposal from proposer.

        Args:
            options (dict): The options for proposer analysis.

        Examples:
            >>> proposer_type = 'Parallel'
            >>> proposer = ProposerFactory.instance().get_proposer(proposer_type, self.profiling_dir, self.rank_id)
            >>> result = proposer.analyze(options)
        """
        if self._parallel_analyser.parallel_mode == self._DATA_PARALLEL:
            self._data_model_parallel_analyze()
        elif self._parallel_analyser.parallel_mode == self._MODEL_PARALLEL:
            self._data_model_parallel_analyze()
        elif self._parallel_analyser.parallel_mode == self._PIPELINE_PARALLEL:
            self._pipeline_parallel_analyze()
        self._clear_warning_node()

        return self._proposal_dict

    def _data_model_parallel_analyze(self):
        """The data-parallel and model-parallel analyze."""
        self._step_interval_analyze()
        self._flops_per_second_analyze()
        self._rank_link_analyze()
        self._communication_analyze()

    def _pipeline_parallel_analyze(self):
        """The pipeline-parallel analyze."""
        self._step_interval_analyze()
        self._flops_per_second_analyze()
        self._rank_link_analyze()
        self._between_stage_communication_analyze()
        self._within_stage_communication_analyze()
        self._flops_analyze()

    def _step_interval_analyze(self):
        """The step interval analyze, find the device with a long step time."""
        if not self._analyse_op_segmentation:
            return

        step_intervals = self._parallel_analyser.get_step_interval()
        avg_step_interval = sum([s["step_interval"] for s in step_intervals]) / len(step_intervals)
        rank_id = -1
        val = 0.0
        percent = 0.0
        for step_interval in step_intervals:
            # get how much more than the average
            if avg_step_interval == 0:
                proportion = 0.0
            else:
                proportion = step_interval["step_interval"] / avg_step_interval - 1
            if proportion > self._step_interval_threshold and proportion > percent:
                rank_id = step_interval["rank_id"]
                val = step_interval["step_interval"]
                percent = proportion
        if rank_id >= 0:
            self._proposal_dict["step_interval"]["rank_id"] = rank_id
            self._proposal_dict["step_interval"]["val"] = round(val, self._VAL_DECIMAL_PLACES)
            self._proposal_dict["step_interval"]["percent"] = round(percent * 100, self._PERCENT_DECIMAL_PLACES)

    def _flops_per_second_analyze(self):
        """The flops per second analyze, find slow rank."""
        if not self._analyse_flops:
            return

        flops_per_seconds = self._parallel_analyser.get_rank_flops_per_second()
        avg_flop = sum([f["FLOPS"] for f in flops_per_seconds]) / len(flops_per_seconds)
        rank_id = -1
        val = 0.0
        percent = 0.0
        for flops in flops_per_seconds:
            proportion = (avg_flop - flops["FLOPS"]) / avg_flop
            if proportion > self._flops_per_second_threshold and proportion > percent:
                rank_id = flops["rank_id"]
                val = flops["FLOPS"]
                percent = proportion
        if rank_id >= 0:
            self._proposal_dict["flops_per_second"]["rank_id"] = rank_id
            self._proposal_dict["flops_per_second"]["val"] = round(val, self._VAL_DECIMAL_PLACES)
            self._proposal_dict["flops_per_second"]["percent"] = round(percent * 100, self._PERCENT_DECIMAL_PLACES)

    def _rank_link_analyze(self):
        """The bandwidth of rank link analyze, find slow rank link."""
        if not self._analyse_hccl:
            return

        rank_links = self._parallel_analyser.get_rank_link_summary()

        # distinguish between RDMA and SDMA
        type_links = {}
        sum_bandwidth = {}
        for rank_link in rank_links:
            src, des = [int(r) for r in rank_link[0].split('-')]
            if src == des:
                continue
            link_type = rank_link[1]
            if link_type not in type_links.keys():
                type_links[link_type] = []
                sum_bandwidth[link_type] = 0
            type_links[link_type].append([src, des, rank_link[4]])
            sum_bandwidth[link_type] = sum_bandwidth[link_type] + rank_link[4]

        warning_link = {}
        for link_type in type_links:
            warning_link[link_type] = []
            avg_bandwidth = sum_bandwidth[link_type] / len(type_links[link_type])
            for link in type_links[link_type]:
                if not avg_bandwidth:
                    continue
                proportion = (avg_bandwidth - link[2]) / avg_bandwidth
                if proportion > self._rank_link_threshold and \
                        (not warning_link[link_type] or proportion > warning_link[link_type][4]):
                    warning_link[link_type] = (link[0], link[1], link_type, link[2], proportion)

        for link in warning_link:
            link_info = warning_link.get(link)
            if link_info:
                self._proposal_dict.get("cluster_link").append({
                    "src_rank": link_info[0],
                    "des_rank": link_info[1],
                    "link_type": link_info[2],
                    "val": round(link_info[3], self._VAL_DECIMAL_PLACES),
                    "percent": round(link_info[4] * 100, self._PERCENT_DECIMAL_PLACES)
                })

    def _communication_analyze(self):
        """The communication time analyze, get the percentage of communication time."""
        if not self._analyse_op_segmentation:
            return

        communication_rate = self._parallel_analyser.get_communication_rate()
        if self._parallel_mode == self._DATA_PARALLEL and \
                communication_rate > self._data_parallel_communication_threshold:
            self._proposal_dict.get("parallel_strategy").get("data_parallel_communication_rate")["val"] = \
                round(communication_rate * 100, self._PERCENT_DECIMAL_PLACES)
            self._proposal_dict.get("parallel_strategy").get("data_parallel_communication_rate")["threshold"] = \
                round(self._data_parallel_communication_threshold * 100, self._PERCENT_DECIMAL_PLACES)
        elif self._parallel_mode == self._MODEL_PARALLEL and \
                communication_rate > self._model_parallel_communication_threshold:
            self._proposal_dict.get("parallel_strategy").get("model_parallel_communication_rate")["val"] = \
                round(communication_rate * 100, self._PERCENT_DECIMAL_PLACES)
            self._proposal_dict.get("parallel_strategy").get("model_parallel_communication_rate")["threshold"] = \
                round(self._model_parallel_communication_threshold * 100, self._PERCENT_DECIMAL_PLACES)

    def _between_stage_communication_analyze(self):
        """The communication time of between stages analyze, get the percentage of communication time."""
        if not self._analyse_op_segmentation:
            return

        communication_rate = self._parallel_analyser.get_between_stage_communication_rate()
        if communication_rate > self._between_stage_communication_threshold:
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_between_stage")["val"] = \
                round(communication_rate * 100, self._PERCENT_DECIMAL_PLACES)
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_between_stage")["threshold"] = \
                round(self._between_stage_communication_threshold * 100, self._PERCENT_DECIMAL_PLACES)

    def _within_stage_communication_analyze(self):
        """The communication time of within stage analyze, get the percentage of communication time."""
        if not self._analyse_op_segmentation:
            return

        communication_rates = self._parallel_analyser.get_within_stage_communication_rate()
        stage_id = -1
        val = 0.0
        for communication_rate in communication_rates:
            if communication_rate['rate'] > self._within_stage_communication_threshold and \
                    communication_rate['rate'] > val:
                stage_id = communication_rate['stage_id']
                val = communication_rate['rate']
        if stage_id > 0:
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_within_stage")["val"] = \
                round(val * 100, self._PERCENT_DECIMAL_PLACES)
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_within_stage")["threshold"] = \
                round(self._within_stage_communication_threshold * 100, self._PERCENT_DECIMAL_PLACES)
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_within_stage")["stage"] = stage_id

    def _flops_analyze(self):
        """The flops of stage analyze, find the stage with the most computation."""
        if not self._analyse_flops:
            return

        rank_flops = self._parallel_analyser.get_rank_flops()

        stage_rank_avg_flops = {}
        stage_rank_size = self._rank_size // self._stage_num
        for rank_flop in rank_flops:
            stage = int(rank_flop["rank_id"] // stage_rank_size + 1)
            stage_rank_avg_flops[stage] = stage_rank_avg_flops.get(stage, 0) + rank_flop["FLOPs"] / stage_rank_size

        stage_avg = sum([stage_rank_avg_flops[stage] for stage in stage_rank_avg_flops]) / len(stage_rank_avg_flops)

        warn_stage = -1
        percent = 0.0
        val = 0.0
        for stage in stage_rank_avg_flops:
            proportion = stage_rank_avg_flops[stage] / stage_avg - 1
            if proportion > self._flops_threshold and proportion > percent:
                warn_stage = stage
                percent = proportion
                val = stage_rank_avg_flops.get(stage)
        if warn_stage >= 0:
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_flops")["stage"] = warn_stage
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_flops")["val"] = \
                round(val, self._VAL_DECIMAL_PLACES)
            self._proposal_dict.get("parallel_strategy").get("pipeline_parallel_flops")["percent"] = \
                round(percent * 100, self._PERCENT_DECIMAL_PLACES)

    def _clear_warning_node(self):
        """Clear useless warning node."""
        for warning in list(self._proposal_dict.get("parallel_strategy").keys()):
            if not self._proposal_dict["parallel_strategy"][warning]:
                del self._proposal_dict.get("parallel_strategy")[warning]

        for warning in list(self._proposal_dict.keys()):
            if not self._proposal_dict[warning]:
                del self._proposal_dict[warning]

    def _validate_parallel_profiler_files(self):
        """validate the parallel files"""
        step_trace_filename = 'step_trace_raw_{rank_id}_detail_time.csv'
        cluster_filename = 'ascend_cluster_analyse_{parallel_mode}_{stage_num}_{rank_size}_{rank_id}.csv'
        flops_filename = 'flops_summary_{rank_id}.json'
        hccl_filename = 'hccl_raw_{rank_id}.csv'

        file_warning = 'parallel performance analysis cannot find the file %s, Please collect all device files.'

        files = os.listdir(self.profiling_path)

        for rank_id in range(self._rank_size):

            file_name = step_trace_filename.format(rank_id=rank_id)
            if file_name not in files and self._analyse_op_segmentation:
                self._analyse_op_segmentation = False
                logger.warning(file_warning, file_name)

            file_name = cluster_filename.format(parallel_mode=self._parallel_mode, stage_num=self._stage_num
                                                , rank_size=self._rank_size, rank_id=rank_id)
            if file_name not in files and self._analyse_op_segmentation:
                self._analyse_op_segmentation = False
                logger.warning(file_warning, file_name)

            file_name = flops_filename.format(rank_id=rank_id)
            if file_name not in files and self._analyse_flops:
                self._analyse_flops = False
                logger.warning(file_warning, file_name)

            file_name = hccl_filename.format(rank_id=rank_id)
            if file_name not in files and self._analyse_hccl:
                self._analyse_hccl = False
                logger.warning(file_warning, file_name)
